from types import SimpleNamespace
import numpy.testing as npt
from fluxpart.partition import fvspart_interval


def test_fvspart_interval():
    """
    References
    ----------
    [PRV14] L Palatella, G Rana, D Vitale (2014), Towards a flux-
    partitioning procedure based on the direct use of high-frequency
    eddy-covariance data.  Boundary-Layer Meteorol. 153:327--337,
    doi:10.1007/s10546-014-9947-x
    """

    # April 7 "physical" example from [PRV14] Table 1
    wue = -37.158598e-3
    qcdata = SimpleNamespace(
        var_q=0.411163e-3 ** 2,
        var_c=5.182580e-6 ** 2,
        wq=0.033140e-3,
        wc=-0.472108e-6,
        corr_qc=-0.881017,
    )
    # Results given in [PRV14]
    desired = SimpleNamespace(
        var_cp=5.230500e-6 ** 2,
        corr_cp_cr=-0.757626,
        Fqt=0.012878e-3,
        Fqe=0.020262e-3,
        Fcp=-0.476489e-6,
        Fcr=0.004381e-6,
    )
    massfluxes, fvsp = fvspart_interval(qcdata, wue)
    assert_partition(massfluxes, fvsp, desired)

    # April 5 "non-physical" example from [PRV14] Table 1
    wue = -24.558131e-3
    qcdata = SimpleNamespace(
        var_q=0.455994e-3 ** 2,
        var_c=4.544450e-6 ** 2,
        wq=0.062700e-3,
        wc=-0.712862e-6,
        corr_qc=-0.922292,
    )
    # Resutls given in [PRV14]
    desired = SimpleNamespace(
        var_cp=3.951510e-6 ** 2,
        corr_cp_cr=-0.724706,
        Fqt=0.025416e-3,
        Fqe=0.037284e-3,
        Fcp=-0.624172e-6,
        Fcr=-0.088690e-6,
    )
    mass_fluxes, fvsp = fvspart_interval(qcdata, wue)
    # ver. 0.2.0 no longer returns nonphysical soln
    if 0:
        assert_partition(mass_fluxes, fvsp, desired)
    assert not fvsp.rootsoln.valid_root

    # comparable to Ray Anderson peach data, 2012-06-07 1300
    # The valid solution in this case uses the '+' CO2 root
    wue = -7.060177e-3
    qcdata = SimpleNamespace(
        var_q=0.40639e-6,
        var_c=7.68505e-12,
        wq=0.1506337e-3,
        wc=-0.6254288e-6,
        corr_qc=-.9501656,
    )
    massfluxes, fvsp = fvspart_interval(qcdata, wue)
    assert fvsp.rootsoln.valid_root

    # TODO: Add a test using the '-' root.
    # Previous test used here was wrong


def assert_partition(fluxes, fvsp, desired):
    if fvsp.rootsoln.valid_root:
        npt.assert_allclose(fvsp.rootsoln.var_cp, desired.var_cp, atol=1e-14)
        npt.assert_allclose(
            fvsp.rootsoln.corr_cp_cr, desired.corr_cp_cr, atol=1e-2
        )
        npt.assert_allclose(fluxes.Fqt, desired.Fqt, atol=1e-7)
        npt.assert_allclose(fluxes.Fqe, desired.Fqe, atol=1e-7)
        npt.assert_allclose(fluxes.Fcp, desired.Fcp, atol=1e-9)
        npt.assert_allclose(fluxes.Fcr, desired.Fcr, atol=1e-10)
    else:
        assert False


if __name__ == "__main__":
    test_fvspart_interval()



"""Physical constants"""

from collections import namedtuple


class MolecularWeight(namedtuple("MolecularWeight", "dryair vapor co2")):
    """
    Attributes
    ----------
    dryair, vapor, co2 : float
        kg/mol
    """


class SpecificGasConstant(
    namedtuple("SpecificGasConstant", "dryair vapor co2")
):
    """
    Attributes
    ----------
    dryair, vapor, co2 : float
        J/K/kg
    """


class SpecificHeatCapacity(namedtuple("SpecificHeatCapacity", "dryair")):
    """
    Attributes
    ----------
    dryair : float
        J/K/kg
    """


GRAVITY = 9.81  # m/s^2
VON_KARMAN = 0.40  # unitless
UNIVERSAL_GAS = 8.3144598  # J/K/mol

MOLECULAR_WEIGHT = MolecularWeight(
    dryair=0.0289645, vapor=0.018016, co2=0.044010
)

SPECIFIC_GAS_CONSTANT = SpecificGasConstant(
    dryair=UNIVERSAL_GAS / MOLECULAR_WEIGHT.dryair,
    vapor=UNIVERSAL_GAS / MOLECULAR_WEIGHT.vapor,
    co2=UNIVERSAL_GAS / MOLECULAR_WEIGHT.co2,
)

SPECIFIC_HEAT_CAPACITY = SpecificHeatCapacity(dryair=1004.67)



from copy import deepcopy
import datetime as pydatetime
from functools import lru_cache
from glob import iglob
import os
import pickle
import zipfile

import attr

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd

from .__version__ import __version__
from .wue import water_use_efficiency, WUEError
from .hfdata import (
    HFData,
    HFSummary,
    HFDataSource,
    HFDataReadError,
    TooFewDataError,
)
from .partition import fvspart_progressive, FVSPSolution
from .util import vapor_press_deficit
from .containers import AllFluxes, WUE
from .constants import MOLECULAR_WEIGHT as MW


EC_TOA5 = {
    "filetype": "csv",
    "skiprows": 4,
    "time_col": 0,
    "cols": (2, 3, 4, 5, 6, 7, 8),
    "temper_unit": "C",
    "unit_convert": dict(q=1e-3, c=1e-6, P=1e3),
    "na_values": "NAN",
}

EC_TOB1 = {
    "filetype": "tob1",
    "time_col": 0,
    "cols": (3, 4, 5, 6, 7, 8, 9),
    "temper_unit": "C",
    "unit_convert": dict(q=1e-3, c=1e-6, P=1e3),
}

EC_GHG1 = {
    "filetype": "ghg",
    "sep": "\t",
    "cols": (11, 12, 13, 7, 8, 9, 10),
    "time_col": [5, 6],
    "unit_convert": dict(q=1e-3 * MW.vapor, c=1e-3 * MW.co2, P=1e3),
    "temper_unit": "C",
    "skiprows": 8,
    "na_values": "NAN",
    "to_datetime_kws": {"format": "%Y-%m-%d %H:%M:%S:%f"},
}

HFD_FORMAT = EC_TOA5

WUE_OPTIONS = {
    "ci_mod": "const_ratio",
    "ci_mod_param": None,
    "leaf_temper": None,
    "leaf_temper_corr": 0,
    "diff_ratio": 1.6,
}

HFD_OPTIONS = {
    "bounds": {"c": (0, np.inf), "q": (0, np.inf)},
    "rd_tol": 0.5,
    "ad_tol": 1024,
    "ustar_tol": 0.1,
    "correct_external": True,
}

PART_OPTIONS = dict(adjust_fluxes=True)

_bad_ustar = "ustar = {:.4} <= ustar_tol = {:.4}"
_bad_vpd = "vpd = {:.4} Pa <= 0"
_bad_qflux = "Fq = {:.4} <= 0"
_night_mssg = "Nighttime, fluxes all non-stomatal"
_fp_result_str = (
    "===============\n"
    "Fluxpart Result\n"
    "===============\n"
    "fluxpart version = {version}\n"
    "date = {date}\n"
    "---------------\n"
    "dataread = {dataread}\n"
    "attempt_partition = {attempt_partition}\n"
    "partition_success = {partition_success}\n"
    "mssg = {mssg}\n"
    "label = {label}\n"
    "sunrise = {sunrise}\n"
    "sunset = {sunset}\n"
    + AllFluxes().results_str()
    + "\n"
    + HFSummary().results_str()
    + "\n"
    + WUE().results_str()
    + "\n"
    + FVSPSolution().results_str()
)


class Error(Exception):
    pass


class FluxpartError(Error):
    def __init__(self, message):
        self.message = message


def fvspart(
    file_or_dir,
    time_sorted=False,
    interval=None,
    hfd_format=None,
    hfd_options=None,
    meas_wue=None,
    wue_options=None,
    part_options=None,
    label=None,
    stdout=True,
    verbose=True,
):
    """Partition CO2 & H2O fluxes into stomatal & nonstomatal components.

    Provides a full implementation of the flux variance similarity
    partitioning algorithm [SS08]_[SAAS+18]_: reads high frequency eddy
    covariance data; performs data transformations and data QA/QC;
    analyzes water vapor and carbon dioxide fluxes; and partitions the
    fluxes into stomatal (transpiration, photosynthesis) and nonstomatal
    (evaporation, respiration) components.

    The following notation is used in variable naming and documentation
    to represent meteorological quantities::

        u, v, w = wind velocities
        q = water vapor mass concentration
        c = carbon dioxide mass concentration
        T = air temperature
        P = total air pressure

    Parameters
    ----------
    For parameters explanation see: :func:`~fluxpart.api.fvs_partition`

    Returns
    -------
    :class:`~fluxpart.fluxpart.FluxpartResult`

    """
    if hfd_format is None:
        hfd_format = deepcopy(HFD_FORMAT)
    elif type(hfd_format) is str and hfd_format.upper() == "EC-TOA5":
        hfd_format = deepcopy(EC_TOA5)
    elif type(hfd_format) is str and hfd_format.upper() == "EC-TOB1":
        hfd_format = deepcopy(EC_TOB1)
    elif type(hfd_format) is str and hfd_format.upper() == "EC-GHG1":
        hfd_format = deepcopy(EC_GHG1)
    else:
        hfd_format = deepcopy(hfd_format)
        _validate_hfd_format(hfd_format)

    hfd_options = {**HFD_OPTIONS, **(hfd_options or {})}
    part_options = {**PART_OPTIONS, **(part_options or {})}

    unit_convert = hfd_format.pop("unit_convert", {})
    converters = {k: _converter_func(v, 0.0) for k, v in unit_convert.items()}
    temper_unit = hfd_format.pop("temper_unit").upper()
    if temper_unit == "C" or temper_unit == "CELSIUS":
        converters["T"] = _converter_func(1.0, 273.15)
    hfd_format["converters"] = converters
    if "daytime" in part_options and type(part_options["daytime"]) is str:
        part_options["daytime"] = _lookup(part_options["daytime"], 0, 1, 2)

    if stdout:
        print("Getting filenames ...")
    files = _files(file_or_dir)
    if len(files) == 0:
        print("No files found!")
        return
    if time_sorted:
        sorted_files = files
    else:
        if stdout:
            print("Reading {} file datetimes ...".format(len(files)))
        times = _peektime(files, **hfd_format)
        if stdout:
            print("Time sorting data files ...")
        sorted_files = [f for t, f in sorted(zip(times, files))]
    if stdout:
        print("Creating data source ...")

    reader = HFDataSource(sorted_files, **hfd_format).reader(interval=interval)
    results = []
    if stdout:
        print("Beginning partitioning analyses ...")

    while True:
        try:
            hfdat = HFData(next(reader))
        except HFDataReadError as e:
            results.append(FVSResult(label=label, mssg=e.args[0]))
            continue
        except StopIteration:
            break

        datetime = hfdat.dataframe.index[0]
        date = datetime.date()
        time = datetime.time()
        if stdout:
            print("{}: ".format(datetime), end="")

        sunrise, sunset = np.nan, np.nan
        nighttime = False
        if "daytime" in part_options:
            if callable(part_options["daytime"]):
                sunrise, sunset = part_options["daytime"](date)
            else:
                sunrise, sunset = part_options["daytime"]
            sunrise = pd.to_datetime(sunrise).time()
            sunset = pd.to_datetime(sunset).time()
            # shift sunrise so we partition if rise occurs during interval
            if interval:
                sunrise = pd.Timestamp.combine(date, sunrise)
                sunrise = (sunrise - pd.Timedelta(interval)).time()
            nighttime = time < sunrise or time > sunset

        try:
            hfdat, hfsum = _set_hfdata(hfdat, **hfd_options)
        except (TooFewDataError, FluxpartError) as e:
            results.append(
                FVSResult(
                    label=datetime,
                    mssg=e.args[0],
                    dataread=True,
                    attempt_partition=False,
                    partition_success=False,
                    sunrise=sunrise,
                    sunset=sunset,
                )
            )
            if stdout and verbose:
                print(e.args[0])
            continue

        if nighttime and hfsum.cov_w_c >= 0:
            fluxes = AllFluxes(
                temper_kelvin=hfsum.T, **_set_all_fluxes_nonstomatal(hfsum)
            )
            mssg = _night_mssg
            results.append(
                FVSResult(
                    label=datetime,
                    attempt_partition=False,
                    partition_success=True,
                    mssg=mssg,
                    dataread=True,
                    fluxes=fluxes,
                    hfsummary=hfsum,
                    sunrise=sunrise,
                    sunset=sunset,
                )
            )
            if stdout and verbose:
                print(mssg)
            continue

        try:
            leaf_wue = _set_leaf_wue(
                meas_wue,
                wue_options,
                part_options,
                hfsum,
                date,
                datetime,
                temper_unit,
            )
        except WUEError as e:
            results.append(
                FVSResult(
                    label=datetime,
                    mssg=e.args[0],
                    dataread=True,
                    hfsummary=hfsum,
                    sunrise=sunrise,
                    sunset=sunset,
                )
            )
            if stdout and verbose:
                print(e.args[0])
            continue

        fluxes, fvsp = fvspart_progressive(
            hfdat["w"].values,
            hfdat["q"].values,
            hfdat["c"].values,
            leaf_wue.wue,
            part_options["adjust_fluxes"],
        )

        if fvsp.valid_partition:
            fluxes = AllFluxes(**attr.asdict(fluxes), temper_kelvin=hfsum.T)
        else:
            fluxes = AllFluxes()

        results.append(
            FVSResult(
                label=datetime,
                sunrise=sunrise,
                sunset=sunset,
                dataread=True,
                attempt_partition=True,
                fluxes=fluxes,
                partition_success=fvsp.valid_partition,
                mssg=fvsp.mssg,
                fvsp_solution=fvsp,
                hfsummary=hfsum,
                wue=leaf_wue,
            )
        )
        if stdout:
            if fvsp.mssg:
                print(fvsp.mssg)
            else:
                print("OK")

    return FluxpartResult(results)


def _set_all_fluxes_nonstomatal(hfsum):
    return dict(
        Fq=hfsum.cov_w_q,
        Fqt=0.0,
        Fqe=hfsum.cov_w_q,
        Fc=hfsum.cov_w_c,
        Fcp=0.0,
        Fcr=hfsum.cov_w_c,
    )


def _set_only_total_fluxes(hfsum):
    return dict(
        Fq=hfsum.cov_w_q,
        Fqt=np.nan,
        Fqe=np.nan,
        Fc=hfsum.cov_w_c,
        Fcp=np.nan,
        Fcr=np.nan,
    )


def _set_hfdata(hfdata, bounds, rd_tol, ad_tol, correct_external, ustar_tol):
    hfdata.cleanse(bounds, rd_tol, ad_tol)
    hfdata.truncate_pow2()
    if correct_external:
        hfdata.correct_external()
    hfsum = hfdata.summarize()
    if hfsum.ustar < ustar_tol:
        raise FluxpartError(_bad_ustar.format(hfsum.ustar, ustar_tol))
    vpd = vapor_press_deficit(hfsum.rho_vapor, hfsum.T)
    if vpd <= 0:
        raise FluxpartError(_bad_vpd.format(vpd))
    if hfsum.cov_w_q <= 0:
        raise FluxpartError(_bad_qflux.format(hfsum.cov_w_q))
    return hfdata, hfsum


def flux_partition(*args, **kws):
    return fvspart(*args, **kws)


class FVSResult(object):
    """FVS partitioning result."""

    def __init__(
        self,
        dataread=False,
        attempt_partition=False,
        partition_success=False,
        mssg=None,
        label=None,
        sunrise=None,
        sunset=None,
        fluxes=AllFluxes(),
        hfsummary=HFSummary(),
        wue=WUE(),
        fvsp_solution=FVSPSolution(),
    ):
        """Fluxpart result.

        Parameters
        ----------
        dataread, attempt_partition, partition_success : bool
            Indicates success or failure in reading high frequency data,
            attempting and obtaining a valid partioning solution.
        mssg : str
            Possibly informative message if `dataread` or `partition_success`
            are False
        label : optional
            Pandas datetime.
        fluxes : :class:`~fluxpart.containers.AllFluxes`
        fvsp_solution : :class:`~fluxpart.containers.FVSPResult`
        wue : :class:`~fluxpart.containers.WUE`
        hfsummary : :class:`~fluxpart.hfdata.HFSummary`

        """
        self.version = __version__
        self.dataread = dataread
        self.attempt_partition = attempt_partition
        self.partition_success = partition_success
        self.mssg = mssg
        self.fluxes = fluxes
        self.label = label
        self.sunrise = sunrise
        self.sunset = sunset
        self.fvsp_solution = fvsp_solution
        self.wue = wue
        self.hfsummary = hfsummary

    def __str__(self):
        fluxpart = attr.asdict(self.fvsp_solution)
        wqc_data = fluxpart.pop("wqc_data")
        rootsoln = fluxpart.pop("rootsoln")
        return _fp_result_str.format(
            timenow=pydatetime.datetime.now(),
            version=self.version,
            dataread=self.dataread,
            attempt_partition=self.attempt_partition,
            partition_success=self.partition_success,
            mssg=self.mssg,
            label=self.label,
            sunrise=self.sunrise,
            sunset=self.sunset,
            **attr.asdict(self.fluxes),
            **attr.asdict(self.hfsummary),
            **attr.asdict(self.wue),
            **fluxpart,
            **wqc_data,
            **rootsoln,
        )


class FluxpartResult(object):
    def __init__(self, fp_results):
        if isinstance(fp_results, pd.DataFrame):
            self.df = fp_results
            return
        index = pd.DatetimeIndex(r.label for r in fp_results)
        df0 = pd.DataFrame(
            (r.fluxes.common_units() for r in fp_results),
            index=index,
            columns=fp_results[0].fluxes.common_units().keys(),
        )
        df1 = pd.DataFrame(
            (r.hfsummary.common_units() for r in fp_results),
            index=index,
            columns=fp_results[0].hfsummary.common_units().keys(),
        )
        df2 = pd.DataFrame(
            (r.wue.common_units() for r in fp_results),
            index=index,
            columns=fp_results[0].wue.common_units().keys(),
        )
        df3 = pd.DataFrame(
            (r.fvsp_solution.common_units() for r in fp_results),
            index=index,
            columns=fp_results[0].fvsp_solution.common_units().keys(),
        )
        df4 = pd.DataFrame(
            {
                "dataread": [r.dataread for r in fp_results],
                "attempt_partition": [r.attempt_partition for r in fp_results],
                "partition_success": [r.partition_success for r in fp_results],
                "mssg": [r.mssg for r in fp_results],
                "sunrise": [r.sunrise for r in fp_results],
                "sunset": [r.sunset for r in fp_results],
            },
            index=index,
        )
        self.df = pd.concat(
            [df0, df1, df2, df3, df4],
            axis=1,
            sort=False,
            keys=["fluxes", "hfsummary", "wue", "fvsp_solution", "fluxpart"],
        )

        self.meta = {
            "version": fp_results[0].version,
            "date": str(pydatetime.datetime.now()),
        }

    def __str__(self):
        if len(self.df) == 1:
            return self.istr(0)
        else:
            return self.df.__str__()

    def __getitem__(self, item):
        return self.df[item]

    def __getattr__(self, x):
        return getattr(self.df, x)

    def plot_co2(
        self,
        start=None,
        end=None,
        units="mass",
        components=(0, 1, 2),
        ax=None,
        **kws,
    ):
        if ax is None:
            ax = plt.gca()
        if units == "mass":
            cols = ["Fc", "Fcp", "Fcr"]
            ylab = r"$\mathrm{CO_2\ Flux\ (mg\ m^{-2}\ s^{-1})}$"
        else:
            cols = ["Fc_mol", "Fcp_mol", "Fcr_mol"]
            ylab = r"$\mathrm{CO_2\ Flux\ (umol\ m^{-2}\ s^{-1})}$"
        labels = [
            r"$\mathrm{F_c}$",
            r"$\mathrm{F_{c_p}}$",
            r"$\mathrm{F_{c_r}}$",
        ]
        cols = [cols[j] for j in components]
        labels = [labels[j] for j in components]
        self.df.loc[start:end, ("fluxes", cols)].plot(ax=ax)
        ax.legend(labels)
        ax.set_ylabel(ylab)
        return ax

    def plot_h2o(
        self,
        start=None,
        end=None,
        units="mass",
        components=(0, 1, 2),
        ax=None,
        **kws,
    ):
        if ax is None:
            ax = plt.gca()
        if units == "mass":
            cols = ["Fq", "Fqt", "Fqe"]
            ylab = r"$\mathrm{H_20\ Flux\ (g\ m^{-2}\ s^{-1})}$"
        elif units == "mol":
            cols = ["Fq_mol", "Fqt_mol", "Fqe_mol"]
            ylab = r"$\mathrm{H_20\ Flux\ (mmol\ m^{-2}\ s^{-1})}$"
        else:
            cols = ["LE", "LEt", "LEe"]
            ylab = r"$\mathrm{LE\ (W\ m^{-2})}$"
        labels = [
            r"$\mathrm{F_q}$",
            r"$\mathrm{F_{q_t}}$",
            r"$\mathrm{F_{q_e}}$",
        ]

        cols = [cols[j] for j in components]
        labels = [labels[j] for j in components]

        self.df.loc[start:end, ("fluxes", cols)].plot(ax=ax)
        ax.legend(labels)
        ax.set_ylabel(ylab)
        return ax

    def istr(self, i):
        """Return a string representation of the ith result"""
        return _fp_result_str.format(
            version=self.meta["version"],
            date=self.meta["date"],
            label=self.df.index[i],
            **self.df.iloc[i]["fluxpart"].to_dict(),
            **self.df.iloc[i]["fluxes"].to_dict(),
            **self.df.iloc[i]["fvsp_solution"].to_dict(),
            **self.df.iloc[i]["hfsummary"].to_dict(),
            **self.df.iloc[i]["wue"].to_dict(),
        )

    def save(self, filename):
        self.save_pickle(filename)

    def save_csv(self, filename):
        self.df.to_csv(filename) #, na_rep="NAN")

    def save_pickle(self, filename):
        self.df.to_pickle(filename)


def _converter_func(slope, intercept):
    """Return a function for linear transform of data."""

    if type(slope) is str:
        return slope

    def func(val):
        return slope * val + intercept

    return func


def _files(file_or_dir):
    if type(file_or_dir) is str:
        file_or_dir = [file_or_dir]
    unsorted_files = []
    for path in file_or_dir:
        if os.path.isfile(path):
            unsorted_files.append(path)
            continue
        if os.path.isdir(path):
            path = os.path.join(path, "*")
        unsorted_files += iglob(path)
    return unsorted_files


def _set_leaf_wue(
    meas_wue, wue_options, part_options, hfsum, date, datetime, temper_unit
):
    # TODO: these should work with file objects, not just str name
    wue_options = {**WUE_OPTIONS, **(wue_options or {})}
    heights, leaf_temper = None, None
    canopy_ht = wue_options.pop("canopy_ht", None)
    meas_ht = wue_options.pop("meas_ht", None)
    if "heights" in wue_options:
        heights = wue_options.pop("heights")
        if not callable(heights):
            heights = _lookup(heights, 0, 1, 2)
    if "leaf_temper" in wue_options:
        leaf_temper = wue_options.pop("leaf_temper")
        if type(leaf_temper) is str:
            leaf_temper = _lookup(leaf_temper, 0, 1)
    if meas_wue:
        if type(meas_wue) is str:
            meas_wue = _lookup(meas_wue, 0, 1)

    try:
        if meas_wue:
            if callable(meas_wue):
                leaf_wue = WUE(wue=meas_wue(datetime))
            else:
                leaf_wue = WUE(wue=float(meas_wue))
        else:
            if heights is not None:
                if callable(heights):
                    canopy_ht, meas_ht = heights(date)
                else:
                    canopy_ht, meas_ht = heights
            else:
                if callable(canopy_ht):
                    canopy_ht = canopy_ht(date)
                if callable(meas_ht):
                    meas_ht = meas_ht(date)
            leaf_t = None
            if leaf_temper is not None:
                if callable(leaf_temper):
                    leaf_t = leaf_temper(datetime)
                else:
                    leaf_t = float(leaf_temper)
                if temper_unit == "C" or temper_unit == "CELSIUS":
                    leaf_t = leaf_t + 273.15
            leaf_wue = water_use_efficiency(
                hfsum,
                canopy_ht=canopy_ht,
                meas_ht=meas_ht,
                leaf_temper=leaf_t,
                **wue_options,
            )
        return leaf_wue

    except WUEError:
        raise


def _peektime(files, **kwargs):
    if kwargs["filetype"] == "csv" or kwargs["filetype"] == "ghg":
        dtcols = kwargs["time_col"]
        if type(dtcols) is int:
            dtcols = [dtcols]
        sep = kwargs.get("delimiter", ",")
        sep = kwargs.get("sep", sep)
        datetimes = []
        to_datetime_kws = kwargs.get("to_datetime_kws", {})
    if kwargs["filetype"] == "csv":
        for file_ in files:
            with open(file_, "rt") as f:
                for _ in range(kwargs.get("skiprows", 0)):
                    f.readline()
                row = f.readline().split(sep)
                tstamp = " ".join([row[i].strip("'\"") for i in dtcols])
                datetimes.append(pd.to_datetime(tstamp, **to_datetime_kws))
    elif kwargs["filetype"] == "ghg":
        for file_ in files:
            with zipfile.ZipFile(file_) as z:
                with z.open(os.path.basename(file_)[:-3] + "data", "r") as f:
                    for _ in range(kwargs.get("skiprows", 0)):
                        f.readline()
                    row = f.readline().decode("utf-8").split(sep)
                    tstamp = " ".join([row[i].strip("'\"") for i in dtcols])
                    datetimes.append(pd.to_datetime(tstamp, **to_datetime_kws))
    else:  # "tob1"
        source = HFDataSource(files, count=5, **kwargs)
        datetimes = [df.index[0] for df in source.reader(interval=None)]
    return datetimes


def _validate_hfd_format(hfd_format):
    if "cols" not in hfd_format:
        raise Error("No value for hfd_format['cols'] given.")
    if "filetype" not in hfd_format:
        raise Error("No value for hfd_format['filetype'] given.")
    if hfd_format["filetype"] not in ("csv", "tob1", "ghg"):
        raise Error(f"Unrecognized filetype: {hfd_format['filetype']}")


def _lookup(csv_file, date_icol, icol1, icol2=None, method="ffill"):
    """Create a function for looking up data in csv file.
    date_icol, icol1, icol2 : int
        column index for the respective data
    method : str
        Interpolation method used with pandas df.index.get_loc. The
        default 'ffill' returns the PREVIOUS values if no exact date
        match is found in the lookup.

    """
    df = pd.read_csv(csv_file, index_col=date_icol, parse_dates=True)

    @lru_cache()
    def func(date):
        ix = df.index.get_loc(pd.to_datetime(date), method=method)
        if icol2 is None:
            return df.iloc[ix, icol1 - 1]
        else:
            return df.iloc[ix, icol1 - 1], df.iloc[ix, icol2 - 1]

    return func



"""
Read and model eddy covariance high-frequency time series data.

The following notation is used in variable naming to represent
meteorological quantities (SI units)::

    u, v, w = wind velocities (m/s)
    q = water vapor mass concentration (kg/m^3)
    c = carbon dioxide mass concentration (kg/m^3)
    T = air temperature (K)
    P = total air pressure (Pa)

"""
import attr
import math
import numpy as np
import pandas as pd

from . import util
from .constants import MOLECULAR_WEIGHT as MW
from .constants import SPECIFIC_HEAT_CAPACITY as CP
from .constants import SPECIFIC_GAS_CONSTANT as GC


class Error(Exception):
    pass


class HFDataReadError(Error):
    def __init__(self, message):
        self.message = message


class TooFewDataError(Error):
    pass


VAR_NAMES = ["u", "v", "w", "c", "q", "T", "P"]

_badfiletype = "File type not recognized ({})"
_toofewdata_rel = "data frac = {frac:.4} < rd_tol = {rd_tol:.4}"
_toofewdata_abs = "data length = {N} < ad_tol = {ad_tol}"


class HFData(object):
    """
    Parameters
    ----------
    hf_dataframe
        High frequency eddy covariance dataframe. Must include columns
        for ["u", "v", "w", "c", "q", "T", "P"]. Normally, dataframe
        should have a datetime index. Dataframe may include also boolean
        mask columns.

    """

    def __init__(self, hf_dataframe):
        self.dataframe = hf_dataframe
        self._already_corrected_external = False

    def __getitem__(self, name):
        """Column-wise get without specifying dataframe attribute"""
        return self.dataframe.loc[:, name]

    def __setitem__(self, name, value):
        """Column-wise set without specifying dataframe attribute"""
        self.dataframe.loc[:, name] = value

    def cleanse(self, bounds=None, rd_tol=0.5, ad_tol=1024):
        """Apply some data QC, remove bad data.

        If problems are found (data not readable, out-of-bounds, or
        flagged), self.dataframe is modified to contain only the longest
        contiguous stretch of good data. An error is raised if the
        resulting data are too few. The criteria for judging 'too few'
        can be specified in both relative and absolute terms: the
        datafile is rejected if the good stretch is a fraction of the
        total data that is less than `rd_tol`, and/or is less than
        `ad_tol` records long.

        Parameters
        ----------
        bounds : dict, optional
            Dictionary specifying lower and upper bounds for legal data.
            Dict entries have the form ``varname: (lower, upper)``,
            where varname is one of 'u', 'v', 'w', 'q', 'c', 'T', or
            'P', and the 2-tuple holds values for the lower and upper
            bounds. Default is None.
        rd_tol : float, optional
            Relative tolerance for rejecting the datafile. Default is
            `rd_tol` = 0.5.
        ad_tol : int, optional
            Absolute tolerance for rejecting the datafile. Default is
            `ad_tol` = 1024.

        """
        bounds = bounds or {}
        data = self.dataframe

        # 1D mask is True for a row if any data are nan, any flag is
        # True, or any data are out-of-bounds
        varindx = pd.Index(VAR_NAMES)
        mask = data.loc[:, varindx].isnull().any(axis=1)
        mask |= data.loc[:, data.columns.difference(varindx)].any(axis=1)
        for var, (low, high) in bounds.items():
            mask |= (data[var] < low) | (data[var] > high)
        if isinstance(data.index, pd.DatetimeIndex) and data.shape[0] > 1:
            diff = data.index.to_series().diff()
            mask |= (abs((diff / diff.mode()[0]) - 1) > .0001)

        # Find longest span of valid (unmasked) data
        marray = np.ma.array(np.zeros([data.shape[0]]), mask=mask.values)
        unmasked_slices = np.ma.clump_unmasked(marray) or [slice(0, 0)]
        max_indx = np.argmax([s.stop - s.start for s in unmasked_slices])
        max_slice = unmasked_slices[max_indx]
        len_max_slice = max_slice.stop - max_slice.start

        # verify sufficient data length
        data_frac = len_max_slice / data.shape[0]
        if data_frac < rd_tol:
            self.dataframe = None
            mssg = _toofewdata_rel.format(frac=data_frac, rd_tol=rd_tol)
            raise TooFewDataError(mssg)
        if len_max_slice < ad_tol:
            self.dataframe = None
            mssg = _toofewdata_abs.format(N=len_max_slice, ad_tol=ad_tol)
            raise TooFewDataError(mssg)

        self.dataframe = data.iloc[max_slice]
        return

    def correct_external(self):
        """Adjust q and c data series to correct for external effects.

        Water vapor and carbon dioxide series data in the dataframe are
        corrected for external fluctuations associated with air
        temperature and vapor density. See: [WPL80]_ and [DK07]_.

        """
        if self._already_corrected_external:
            return
        ave_vapor = self["q"].mean()
        ave_co2 = self["c"].mean()
        ave_T = self["T"].mean()
        dev_vapor = self["q"] - ave_vapor
        dev_T = self["T"] - ave_T

        Pdryair = self["P"].mean() - ave_vapor * GC.vapor * ave_T
        rho_totair = ave_vapor + Pdryair / GC.dryair / ave_T

        specific_vapor = ave_vapor / rho_totair
        specific_co2 = ave_co2 / rho_totair
        mu = MW.dryair / MW.vapor
        muq = mu * specific_vapor
        muc = mu * specific_co2

        self["q"] += muq * dev_vapor + (1 + muq) * ave_vapor * dev_T / ave_T
        self["c"] += muc * dev_vapor + (1 + muq) * ave_co2 * dev_T / ave_T
        self._already_corrected_external = True
        return

    def summarize(self):
        """Summarize high frequency dataframe statistics.

        Returns
        -------
        :class:`~fluxpart.hfdata.HFSummary`

        """
        hfs = util.stats2(self.dataframe, VAR_NAMES)
        Pvap = hfs.ave_q * GC.vapor * hfs.ave_T
        rho_dryair = (hfs.ave_P - Pvap) / GC.dryair / hfs.ave_T
        rho_totair = rho_dryair + hfs.ave_q
        Cp = CP.dryair * (1 + 0.84 * hfs.ave_q / rho_totair)

        return HFSummary(
            T=hfs.ave_T,
            P=hfs.ave_P,
            Pvap=Pvap,
            ustar=(hfs.cov_w_u ** 2 + hfs.cov_w_v ** 2) ** 0.25,
            wind_w=hfs.ave_w,
            var_w=hfs.var_w,
            rho_vapor=hfs.ave_q,
            var_vapor=hfs.var_q,
            rho_co2=hfs.ave_c,
            var_co2=hfs.var_c,
            corr_q_c=hfs.cov_q_c / math.sqrt(hfs.var_q * hfs.var_c),
            H=rho_totair * Cp * hfs.cov_w_T,
            cov_w_q=hfs.cov_w_q,
            cov_w_c=hfs.cov_w_c,
            rho_dryair=rho_dryair,
            rho_totair=rho_totair,
            cov_w_T=hfs.cov_w_T,
            N=self.dataframe.shape[0],
        )

    def truncate_pow2(self):
        """Truncate dataframe length to largest possible power of 2."""
        truncate_len = 2 ** int(np.log2(self.dataframe.shape[0]))
        self.dataframe = self.dataframe.iloc[:truncate_len]


class HFDataSource(object):
    """Reader for high-frequency eddy covariance data.

    Parameters
    ----------
    files : list or sequence of files
        Sorted sequence of data files
    filetype : {'csv', 'tob1', 'ghg'}
        'csv' = delimited text file; 'tob1' = Campbell Scientific binary
        format file; 'ghg' = LI-COR raw data format 
    cols : 7*(int,)
        Column indices for (u, v, w, q, c, T, P) data, in that order.
        0-based indexing.
    converters : dict, required if datafile uses any non-SI units
        Dictionary of functions used to convert any non-SI data to
        SI units.  Dict keys are 'u', 'v', 'w', 'q', 'c', 'T', or
        'P'. Functions take a single argument, e.g.
        ``converters={'P': lambda arg: 1e3 * arg}``. If all data are
        SI units, set `converters` to None (default).
    time_col : int, optional
        Datetime column for `csv` and `ghg` filetype. Default is None.
    flags : 2-tuple or list of 2-tuples, optional
        Indicate that one or more data columns are used to flag bad data
        records. Each tuple is of the form (col, goodval), where col is
        an int specifying the column number (0-based indexing), and
        goodval is the flag value indicating a good data record.
    **kwargs
        Passed to pandas.read_csv_ when filetype is csv or ghg. Should
        not include `usecols` or `header` keywords.


    .. _pandas.read_csv:
        https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html

    """

    def __init__(
        self,
        files,
        filetype,
        cols,
        converters=None,
        time_col=None,
        flags=None,
        **kwargs,
    ):
        if flags is None:
            flags = []
        elif not isinstance(flags, list):
            flags = [flags]

        self._files = files
        self._filetype = filetype.strip().lower()
        self._cols = cols
        self._converters = {} if converters is None else converters
        self._time_col = time_col
        self._flags = flags
        self._csv_kws = kwargs
        self._dt_kws = self._csv_kws.pop("to_datetime_kws", {})

    @property
    def _names(self):
        # sorted because pd.read_csv sorts usecols but not names
        return sorted(self._namecols, key=self._namecols.get)

    @property
    def _namecols(self):
        namecols = dict(zip(VAR_NAMES, self._cols))
        flags = {"flag-" + str(col): col for col, val in self._flags}
        namecols.update(flags)
        if self._filetype in ("csv", "ghg") and self._time_col is not None:
            try:
                namecols["Date"] = self._time_col[0]
                namecols["Time"] = self._time_col[1]
            except TypeError:
                namecols["Datetime"] = self._time_col
        return namecols

    def reader(self, interval, **kwargs):
        """Consume data source and yield in chunks of the time interval.

        Data chunks are returned in dataframe format.

        Parameters
        ----------
        interval : str
            Time interval used to chunk the data. Is specified using the
            pandas string alias_ format (e.g., ``interval="30min"``).
            If set to -1, a single dataframe corresponding to a
            concatenation of all data files is returned on the first
            iteration. If set to None, dataframes corresponding to whole
            individual data files are returned with each iteration.
        **kwargs
            For csv and ghg filetypes, kwargs are passed to
            pandas.read_csv_.  Can be used to override or add to
            formatting specified in the initializer. Should not include
            `usecols`  or `header` keywords. For tob1 filetype, kwargs
            is passed to numpy.fromfile_. In this case, the only
            allowable kwarg is 'count', which can be used to limit the
            number of lines read.


        .. _numpy.fromfile:
            https://docs.scipy.org/doc/numpy/reference/generated/numpy.fromfile.html

        .. _pandas.read_csv:
            https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html

        .. _alias:
            http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

        """

        if self._filetype in ("csv", "ghg"):
            kws = {**self._csv_kws, **kwargs}
            kws["usecols"] = self._usecols
            kws["header"] = None
            if self._filetype == "csv":
                dfs = util.multifile_read_csv(self._files, **kws)
            else:
                dfs = util.multifile_read_ghg(self._files, **kws)
            indx_dfs = (self._set_indices_csv(df) for df in dfs)
        else:  # "tob1"
            count = kwargs.get("count", -1)
            dfs = util.multifile_read_tob1(self._files, count=count)
            indx_dfs = (self._set_indices_tob1(df) for df in dfs)

        si_dfs = (self._set_units(df) for df in indx_dfs)
        hf_dfs = (self._set_flags(df) for df in si_dfs)
        try:
            yield from util.chunked_df(hf_dfs, interval)
        except StopIteration:
            return

    def _set_indices_csv(self, df):
        df.columns = self._names
        if self._time_col is None:
            return df
        try:
            df["Datetime"] = pd.to_datetime(
                df.loc[:, "Datetime"], **self._dt_kws
            )
        except KeyError:
            df["Datetime"] = pd.to_datetime(
                df.loc[:, "Date"] + " " + df.loc[:, "Time"], **self._dt_kws
            )
            df = df.drop(["Date", "Time"], axis=1)
        return df.set_index("Datetime")

    def _set_indices_tob1(self, df):
        df["Datetime"] = pd.to_datetime(
            arg=df.loc[:, "SECONDS"] + 10 ** -9 * df.loc[:, "NANOSECONDS"],
            unit="s",
            origin="1990-01-01",
        )
        df = df.set_index("Datetime")
        df = df.iloc[:, self._usecols]
        df.columns = self._names
        return df

    def _set_flags(self, df):
        for col, val in self._flags:
            df.loc[:, "flag-" + str(col)] = (
                df.loc[:, "flag-" + str(col)] != val
            )
        return df

    def _set_units(self, df):
        conc_are_mole_ratios = False
        for var, func in self._converters.items():
            if callable(func):
                df.loc[:, var] = func(df.loc[:, var])
            else:
                conc_are_mole_ratios = True
        if conc_are_mole_ratios:
            q_units = self._converters["q"]
            c_units = self._converters["c"]
            p = df.loc[:, "P"]
            t = df.loc[:, "T"]
            coef = 1e3 if q_units[:3] == "ppt" else 1e6
            q_mole_ratio = df.loc[:, "q"] / coef
            df.loc[:, "q"] = q_mole_ratio * p / GC.vapor / t
            coef = 1e3 if c_units[:3] == "ppt" else 1e6
            c_mole_ratio = df.loc[:, "c"] / coef
            df.loc[:, "c"] = c_mole_ratio * p / GC.co2 / t
            if c_units[-3:] == "dry":
                df.loc[:, "c"] /= 1 + q_mole_ratio
            if q_units[-3:] == "dry":
                df.loc[:, "q"] /= 1 + q_mole_ratio
        return df

    @property
    def _usecols(self):
        return [self._namecols[k] for k in self._names]


@attr.s
class HFSummary(object):
    """Summary of high frequency eddy covariance data.

    Parameters
    ----------
    T : float
        Mean air temperature, K.
    P, Pvap : float
        Mean total atmospheric (`P`) and vapor (`Pvap`) pressure, Pa.
    ustar : float
        Mean friction velocity, m/s.
    wind_w : float
        Mean vertical wind velocity, m/s.
    var_w : float
        Variance of vertical wind velocity, (m/s)^2.
    rho_vapor, rho_co2 : float
        Mean H2O vapor and CO2 concentrations, kg/m^3.
    var_vapor, var_co2 : float
        Variance of H2O vapor and CO2 concentrations, (kg/m^3)^2.
    corr_q_c : float
        Correlation coefficient for H2O and CO2 concentrations
    cov_w_q, cov_w_c : float
        Covariance of vertical wind velocity (w) with water vapor (q)
        and CO2 (c) mass densities, kg/m^2/s.
    H : float
        Sensible heat flux, W/m^2.
    rho_dryair, rho_totair : float
        Dry and moist air densities, kg/m^3.
    cov_w_T : float
        Covariance of temperature and vertical wind velocity, K m/s.
    N : int
        Length of data series.

    """

    T = attr.ib(default=np.nan)
    P = attr.ib(default=np.nan)
    Pvap = attr.ib(default=np.nan)
    ustar = attr.ib(default=np.nan)
    wind_w = attr.ib(default=np.nan)
    var_w = attr.ib(default=np.nan)
    rho_vapor = attr.ib(default=np.nan)
    rho_co2 = attr.ib(default=np.nan)
    var_vapor = attr.ib(default=np.nan)
    var_co2 = attr.ib(default=np.nan)
    corr_q_c = attr.ib(default=np.nan)
    cov_w_q = attr.ib(default=np.nan)
    cov_w_c = attr.ib(default=np.nan)
    H = attr.ib(default=np.nan)
    rho_dryair = attr.ib(default=np.nan)
    rho_totair = attr.ib(default=np.nan)
    cov_w_T = attr.ib(default=np.nan)
    N = attr.ib(default=np.nan)

    def __str__(self):
        # prints common units instead of SI
        return self.results_str().format(**self.common_units())

    @property
    def fc_ov_fq(self):
        return self.cov_w_c / self.cov_w_q

    @property
    def sigc_ov_sigq(self):
        return math.sqrt(self.var_co2 / self.var_vapor)

    def results_str(self):
        lab = self.common_units_labels()
        return (
            "---------------\n"
            "HF Data Summary\n"
            "---------------\n"
            "  T = {T:.4} " + lab["T"] + "\n"
            "  P = {P:.4} " + lab["P"] + "\n"
            "  Pvap = {Pvap:.4} " + lab["Pvap"] + "\n"
            "  ustar = {ustar:.4} " + lab["ustar"] + "\n"
            "  wind_w = {wind_w:.4} " + lab["wind_w"] + "\n"
            "  var_w = {var_w:.4} " + lab["var_w"] + "\n"
            "  rho_vapor = {rho_vapor:.4} " + lab["rho_vapor"] + "\n"
            "  rho_co2 = {rho_co2:.4} " + lab["rho_co2"] + "\n"
            "  var_vapor = {var_vapor:.4} " + lab["var_vapor"] + "\n"
            "  var_co2 = {var_co2:.4} " + lab["var_co2"] + "\n"
            "  corr_q_c = {corr_q_c:.4} " + lab["corr_q_c"] + "\n"
            "  cov_w_q = {cov_w_q:.4} " + lab["cov_w_q"] + "\n"
            "  H = {H:.4} " + lab["H"] + "\n"
            "  cov_w_c = {cov_w_c:.4} " + lab["cov_w_c"] + "\n"
            "  rho_dryair = {rho_dryair:.4} " + lab["rho_dryair"] + "\n"
            "  rho_totair = {rho_totair:.4} " + lab["rho_totair"] + "\n"
            "  cov_w_T = {cov_w_T:.4} " + lab["cov_w_T"] + "\n"
            "  N = {N:.0f} " + lab["N"]
        )

    def common_units(self):
        return dict(
            T=self.T - 273.15,
            P=1e-3 * self.P,
            Pvap=1e-3 * self.Pvap,
            ustar=self.ustar,
            wind_w=self.wind_w,
            var_w=self.var_w,
            rho_vapor=1e3 * self.rho_vapor,
            rho_co2=1e6 * self.rho_co2,
            var_vapor=1e6 * self.var_vapor,
            var_co2=1e12 * self.var_co2,
            corr_q_c=self.corr_q_c,
            cov_w_q=1e3 * self.cov_w_q,
            fc_ov_fq=1e3 * self.fc_ov_fq,
            sigc_ov_sigq=1e3 * self.sigc_ov_sigq,
            H=self.H,
            cov_w_c=1e6 * self.cov_w_c,
            rho_dryair=self.rho_dryair,
            rho_totair=self.rho_totair,
            cov_w_T=self.cov_w_T,
            N=self.N,
        )

    def common_units_labels(self):
        return dict(
            T="C",
            P="kPa",
            Pvap="kPa",
            ustar="m/s",
            wind_w="m/s",
            var_w="(m/s)^2",
            rho_vapor="g/m^3",
            rho_co2="mg/m^3",
            var_vapor="(g/m^3)^2",
            var_co2="(mg/m^3)^2",
            corr_q_c="",
            cov_w_q="g/m^2/s",
            fc_ov_fq="mg/g",
            sigc_ov_sigq="mg/g",
            H="W/m^2",
            cov_w_c="mg/m^2/s",
            rho_dryair="kg/m^3",
            rho_totair="kg/m^3",
            cov_w_T="C m/s",
            N="",
        )



"""Classes used to group data and results."""

import attr
import numpy as np

from .util import qflux_mass_to_heat, qflux_mass_to_mol, cflux_mass_to_mol


@attr.s
class MassFluxes(object):
    """H2O and CO2 mass flux components.

    Parameters
    ----------
    Fq, Fqt, Fqe : float
        Total, transpiration, and evaporation H2O fluxes, kg/m^2/s
    Fc, Fcp, Fcr : float
        Total, photosynthesis, and respiration CO2 fluxes, kg/m^2/s

    """

    Fq = attr.ib(default=np.nan)
    Fqt = attr.ib(default=np.nan)
    Fqe = attr.ib(default=np.nan)
    Fc = attr.ib(default=np.nan)
    Fcp = attr.ib(default=np.nan)
    Fcr = attr.ib(default=np.nan)

    def __str__(self):
        # Print with common mass units instead of SI
        fqs = [1e3 * self.Fq, 1e3 * self.Fqt, 1e3 * self.Fqe]
        fcs = [1e6 * self.Fc, 1e6 * self.Fcp, 1e6 * self.Fcr]
        return (
            "MassFluxes(\n"
            "    Fq = {:.4} g/m^2/s,\n"
            "    Fqt = {:.4} g/m^2/s,\n"
            "    Fqe = {:.4} g/m^2/s,\n"
            "    Fc = {:.4} mg/m^2/s,\n"
            "    Fcp = {:.4} mg/m^2/s,\n"
            "    Fcr = {:.4} mg/m^2/s)"
            "".format(*(fqs + fcs))
        )


@attr.s
class AllFluxes(object):
    """Water vapor and CO2 fluxes.

    Parameters
    ----------
    Fq, Fqt, Fqe : float
        Total, transpiration, and evaporation H2O fluxes, kg/m^2/s.
    Fc, Fcp, Fcr : float
        Total, photosynthesis, and respiration CO2 fluxes, kg/m^2/s.
    temper_kelvin : float
        Temperature, K

    """

    Fq = attr.ib(default=np.nan)
    Fqt = attr.ib(default=np.nan)
    Fqe = attr.ib(default=np.nan)
    Fc = attr.ib(default=np.nan)
    Fcp = attr.ib(default=np.nan)
    Fcr = attr.ib(default=np.nan)
    temper_kelvin = attr.ib(default=np.nan)
    LE = attr.ib(default=False)
    LEt = attr.ib(default=False)
    LEe = attr.ib(default=False)
    Fq_mol = attr.ib(default=False)
    Fqt_mol = attr.ib(default=False)
    Fqe_mol = attr.ib(default=False)
    Fc_mol = attr.ib(default=False)
    Fcp_mol = attr.ib(default=False)
    Fcr_mol = attr.ib(default=False)

    def __attrs_post_init__(self):
        """Water vapor and CO2 fluxes.

        Derived Parameters
        ------------------
        LE, LEt, LEe : float
            Water vapor fluxes expressed as latent heat, W/m^2.
        Fq_mol, Fqt_mol, Fqe_mol : float
            Water vapor fluxes expressed as mol/m^2/s.
        Fc_mol, Fcp_mol, Fcr_mol : float
            CO2 fluxes expressed as mol/m^2/s.

        """
        if self.LE is False:
            self.LE = qflux_mass_to_heat(self.Fq, self.temper_kelvin)
        if self.LEt is False:
            self.LEt = qflux_mass_to_heat(self.Fqt, self.temper_kelvin)
        if self.LEe is False:
            self.LEe = qflux_mass_to_heat(self.Fqe, self.temper_kelvin)
        if self.Fq_mol is False:
            self.Fq_mol = qflux_mass_to_mol(self.Fq)
        if self.Fqt_mol is False:
            self.Fqt_mol = qflux_mass_to_mol(self.Fqt)
        if self.Fqe_mol is False:
            self.Fqe_mol = qflux_mass_to_mol(self.Fqe)
        if self.Fc_mol is False:
            self.Fc_mol = cflux_mass_to_mol(self.Fc)
        if self.Fcp_mol is False:
            self.Fcp_mol = cflux_mass_to_mol(self.Fcp)
        if self.Fcr_mol is False:
            self.Fcr_mol = cflux_mass_to_mol(self.Fcr)

    def __str__(self):
        # print common units instead of SI
        return self.results_str().format(**self.common_units())

    def results_str(self):
        lab = self.common_units_labels()
        return (
            "------\n"
            "Fluxes\n" + "------\n"
            "  Fq = {Fq:.4} " + lab["Fq"] + "\n"
            "  Fqt = {Fqt:.4} " + lab["Fqt"] + "\n"
            "  Fqe = {Fqe:.4} " + lab["Fqe"] + "\n"
            "  Fc = {Fc:.4} " + lab["Fc"] + "\n"
            "  Fcp = {Fcp:.4} " + lab["Fcp"] + "\n"
            "  Fcr = {Fcr:.4} " + lab["Fcr"] + "\n"
            "  Fq_mol = {Fq_mol:.4} " + lab["Fq_mol"] + "\n"
            "  Fqt_mol = {Fqt_mol:.4} " + lab["Fqt_mol"] + "\n"
            "  Fqe_mol = {Fqe_mol:.4} " + lab["Fqe_mol"] + "\n"
            "  Fc_mol = {Fc_mol:.4} " + lab["Fc_mol"] + "\n"
            "  Fcp_mol = {Fcp_mol:.4} " + lab["Fcp_mol"] + "\n"
            "  Fcr_mol = {Fcr_mol:.4} " + lab["Fcr_mol"] + "\n"
            "  LE  = {LE:.4} " + lab["LE"] + "\n"
            "  LEt = {LEt:.4} " + lab["LEt"] + "\n"
            "  LEe = {LEe:.4} " + lab["LEe"]
        )

    def common_units(self):
        return dict(
            Fq=1e3 * self.Fq,
            Fqt=1e3 * self.Fqt,
            Fqe=1e3 * self.Fqe,
            Fc=1e6 * self.Fc,
            Fcp=1e6 * self.Fcp,
            Fcr=1e6 * self.Fcr,
            Fq_mol=1e3 * self.Fq_mol,
            Fqt_mol=1e3 * self.Fqt_mol,
            Fqe_mol=1e3 * self.Fqe_mol,
            Fc_mol=1e6 * self.Fc_mol,
            Fcp_mol=1e6 * self.Fcp_mol,
            Fcr_mol=1e6 * self.Fcr_mol,
            LE=self.LE,
            LEt=self.LEt,
            LEe=self.LEe,
        )

    def common_units_labels(self):
        return dict(
            Fq="g/m^2/s",
            Fqt="g/m^2/s",
            Fqe="g/m^2/s",
            Fc="mg/m^2/s",
            Fcp="mg/m^2/s",
            Fcr="mg/m^2/s",
            Fq_mol="mmol/m^2/s",
            Fqt_mol="mmol/m^2/s",
            Fqe_mol="mmol/m^2/s",
            Fc_mol="umol/m^2/s",
            Fcp_mol="umol/m^2/s",
            Fcr_mol="umol/m^2/s",
            LE="W/m^2",
            LEt="W/m^2",
            LEe="W/m^2",
        )


@attr.s
class RootSoln(object):
    """Results from calculating the root (corr_cp_cr, var_cp).

    Attributes
    ----------
    corr_cp_cr : float
        Correlation coefficient for CO2 concentrations connected with
        photosynthesis (cp) and respiration (cr).
    var_cp : float
        Variance of CO2 concentration connected with photosynthesis,
        (kg/m^3)^2.
    sig_cr : float
        Standard deviation of CO2 concentration connected with
        respiration, kg/m^3.
    co2soln_id : {0 or 1}
        Indicates the solution used for the quadratic Eq. 13b of
        [SS08]_. Equal to 1 for the '+' root, equal to 0 for the '-'
        root.
    valid_root : bool
        Indicates whether the obtained root (`corr_cp_cr`, `var_cp`) is
        physically plausible.
    root_mssg : str
        Possibly informative message if `isvalid` = False.

    """

    corr_cp_cr = attr.ib(default=np.nan)
    var_cp = attr.ib(default=np.nan)
    sig_cr = attr.ib(default=np.nan)
    co2soln_id = attr.ib(default=np.nan)
    valid_root = attr.ib(default=np.nan)

    def __str__(self, head=True):
        # Print common units instead of SI
        return self.results_str(head).format(**self.common_units())

    def results_str(self, head=True):
        lab = self.common_units_labels()
        out = ""
        if head is True:
            out += " ------------\nRoot Solution\n-------------\n"
        out += (
            "  corr_cp_cr = {corr_cp_cr:.4}\n"
            "  var_cp = {var_cp:.4} " + lab["var_cp"] + "\n"
            "  sig_cr = {sig_cr:.4} " + lab["sig_cr"] + "\n"
            "  co2soln_id = {co2soln_id:.0f}\n"
            "  valid_root = {valid_root}\n"
        )
        return out

    def common_units(self):
        return dict(
            corr_cp_cr=self.corr_cp_cr,
            var_cp=1e12 * self.var_cp,
            sig_cr=1e6 * self.sig_cr,
            co2soln_id=self.co2soln_id,
            valid_root=self.valid_root,
        )

    def common_units_labels(self):
        return dict(
            corr_cp_cr="",
            var_cp="(mg/m^3)^2",
            sig_cr="mg/m^3",
            co2soln_id="",
            valid_root="",
        )


@attr.s
class WQCData(object):
    """Summary stats for wind, water vapor, and CO2 data.

    Attributes
    ----------
    var_q, var_c : float
        Variance of vapor (q) and CO2 (c) concentrations, (kg/m^3)^2.
    corr_qc : float
        Correlation coefficient for vapor and CO2 concentrations.
    wq, wc : float
        Mean vapor (`wq`) and CO2 (`wc`) fluxes, kg/m^2/s.

    """

    var_q = attr.ib(default=np.nan)
    var_c = attr.ib(default=np.nan)
    corr_qc = attr.ib(default=np.nan)
    wq = attr.ib(default=np.nan)
    wc = attr.ib(default=np.nan)

    def __str__(self, head=True):
        # Print common units instead of SI
        return self.results_str(head).format(**self.common_units())

    def results_str(self, head=True):
        lab = self.common_units_labels()
        out = ""
        if head:
            out += "--------------\nInterval Stats\n" "--------------\n"
        out += (
            "  var_q = {var_q:.4} " + lab["var_q"] + "\n"
            "  var_c = {var_c:.4} " + lab["var_c"] + "\n"
            "  corr_qc = {corr_qc:.4}\n"
            "  cov_w_q = {wq:.4} " + lab["cov_w_q"] + "\n"
            "  cov_w_c = {wc:.4} " + lab["cov_w_c"]
        )
        return out

    def common_units(self):
        return dict(
            var_q=1e6 * self.var_q,
            var_c=1e12 * self.var_c,
            corr_qc=self.corr_qc,
            wq=1e3 * self.wq,
            wc=1e6 * self.wc,
        )

    def common_units_labels(self):
        return dict(
            var_q="(g/m^3)^2",
            var_c="(mg/m^3)^2",
            corr_qc="",
            cov_w_q="g/m^2/s",
            cov_w_c="mg/m^2/s",
        )


@attr.s
class FVSPSolution(object):
    """Result of FVS partitioning.

    Parameters
    ----------
     wave_lvl : (int, int)
         2-tuple indicating the level of filtering applied (number of
         components removed from the series data). The second int is the
         maximum possible wavelet decompostion level given the length of
         the data. The first is the number of components remaining in the
         data.  So when the first number is equal to the second, no
         components have been removed (no filtering applied). When the
         first number is 1, the maximum level of filtering was applied.

    """

    wqc_data = attr.ib(default=WQCData())
    rootsoln = attr.ib(default=RootSoln())
    wave_lvl = attr.ib(default=np.nan)
    valid_partition = attr.ib(default=np.nan)
    fvsp_mssg = attr.ib(default="")

    def __str__(self):
        return self.results_str().format(
            fvsp_mssg=self.fvsp_mssg,
            valid_partition=self.valid_partition,
            wave_lvl=self.wave_lvl,
            **self.wqc_data.common_units(),
            **self.rootsoln.common_units(),
        )

    def results_str(self):
        result = (
            "-------------\n"
            "FVSP Solution\n"
            "-------------\n"
            "  valid_partition: {valid_partition}\n"
            "  fvsp_mssg: {fvsp_mssg}\n"
        )
        result += "  Wavelet filtering level: {wave_lvl}\n"
        result += self.wqc_data.results_str(head=False) + "\n"
        result += self.rootsoln.results_str(head=False)
        return result

    def common_units(self):
        return {
            "valid_partition": self.valid_partition,
            "fvsp_mssg": self.fvsp_mssg,
            "wave_lvl": self.wave_lvl,
            **self.wqc_data.common_units(),
            **self.rootsoln.common_units(),
        }


@attr.s
class WUE:
    """Summary of leaf-level water use efficiency calculation.

    Attributes
    ----------
    wue : float
        Leaf-level water use efficiency, kg CO2 / kg H2O.
    inter_h2o : float
        Concentration of intercellular water vapor, kg/m^3
    inter_co2 : float
        Concentrations of intercellular CO2, kg/m^3.
    ambient_h2o : float
        Concentrations of ambient atmospheric water vapor, kg/m^3.
    ambient_co2 : float
        Concentration of ambient CO2, kg/m^3.
    vpd : float
        Atmospheric vapor pressure deficit, Pa.
    ci_mod : str
        The name of the model used to estimate the intercellular CO2
        concentration.
    ci_mod_param : float or 2-tuple of floats
        Specific paramter values used with `ci_mod`.
    leaf_temper : float
    ppath : {'C3' or 'C4'}
        Photosynthetic pathway.
    meas_ht, canopy_ht : float
        Eddy covariance measument height and plant canopy height, m.

    """

    wue = attr.ib(default=np.nan)
    inter_h2o = attr.ib(default=np.nan)
    inter_co2 = attr.ib(default=np.nan)
    ambient_h2o = attr.ib(default=np.nan)
    ambient_co2 = attr.ib(default=np.nan)
    vpd = attr.ib(default=np.nan)
    ci_mod = attr.ib(default=np.nan)
    ci_mod_param = attr.ib(default=np.nan)
    leaf_temper = attr.ib(default=np.nan)
    ppath = attr.ib(default=np.nan)
    meas_ht = attr.ib(default=np.nan)
    canopy_ht = attr.ib(default=np.nan)
    diff_ratio = attr.ib(default=np.nan)

    def __str__(self):
        # Print common units instead of SI
        return self.results_str().format(**self.common_units())

    def results_str(self):
        lab = self.common_units_labels()
        return (
            "--------------------\n"
            "Water Use Efficiency\n"
            "--------------------\n"
            "  wue = {wue:.4} " + lab["wue"] + "\n"
            "  inter_h2o = {inter_h2o:.4} " + lab["inter_h2o"] + "\n"
            "  inter_co2 = {inter_co2:.4} " + lab["inter_co2"] + "\n"
            "  ambient_h2o = {ambient_h2o:.4} " + lab["ambient_h2o"] + "\n"
            "  ambient_co2 = {ambient_co2:.4} " + lab["ambient_co2"] + "\n"
            "  vpd = {vpd:.4} " + lab["vpd"] + "\n"
            "  ci_mod = {ci_mod} " + lab["ci_mod"] + "\n"
            "  ci_mod_param = {ci_mod_param} " + lab["ci_mod_param"] + "\n"
            "  leaf_temper = {leaf_temper:.4} " + lab["leaf_temper"] + "\n"
            "  ppath = {ppath} " + lab["ppath"] + "\n"
            "  meas_ht = {meas_ht} " + lab["meas_ht"] + "\n"
            "  canopy_ht = {canopy_ht} " + lab["canopy_ht"] + "\n"
            "  diff_ratio = {diff_ratio:.4} " + lab["diff_ratio"]
        )

    def common_units(self):
        return dict(
            wue=1e3 * self.wue,
            inter_h2o=1e3 * self.inter_h2o,
            inter_co2=1e6 * self.inter_co2,
            ambient_h2o=1e3 * self.ambient_h2o,
            ambient_co2=1e6 * self.ambient_co2,
            vpd=1e-3 * self.vpd,
            ci_mod=self.ci_mod,
            ci_mod_param=self.ci_mod_param,
            leaf_temper=self.leaf_temper - 273.15,
            ppath=self.ppath,
            meas_ht=self.meas_ht,
            canopy_ht=self.canopy_ht,
            diff_ratio=self.diff_ratio,
        )

    def common_units_labels(self):
        return dict(
            wue="mg/g",
            inter_h2o="g/m^3",
            inter_co2="mg/m^3",
            ambient_h2o="g/m^3",
            ambient_co2="mg/m^3",
            vpd="kPa",
            ci_mod="",
            ci_mod_param="",  # depends on the mod
            leaf_temper="C",
            ppath="",
            meas_ht="m",
            canopy_ht="m",
            diff_ratio="",
        )



from collections import namedtuple
from itertools import permutations
from math import exp
import os
import warnings
import zipfile

import numpy as np
import pandas as pd
import pywt

from .constants import MOLECULAR_WEIGHT as MW  # kg/mol
from .constants import SPECIFIC_GAS_CONSTANT as Rgas  # m^3 Pa / (kg K)


NP_TYPE = {
    "IEEE4": np.float32,
    "IEEE8": np.float64,
    "LONG": np.int32,
    "ULONG": np.uint32,
}


class HFDataReadWarning(UserWarning):
    pass


def chunked_df(dataframes, time_interval):
    """Partition time-indexed dataframe sequence into time intervals."""

    if time_interval == -1:
        yield pd.concat(dataframes)
        return
    if time_interval is None:
        for df in dataframes:
            yield df
        return

    next_df = next(dataframes)
    consumed_dfs = [next_df]
    current_interval = next_df.index[0].floor(time_interval)
    for next_df in dataframes:
        consumed_dfs.append(next_df)
        if next_df.index[-1].floor(time_interval) == current_interval:
            continue

        # If we reach here, then at least one new interval has been
        # found. We yield sequentially all read intervals except the
        # last one because more of that interval could still be
        # coming

        df = pd.concat(consumed_dfs)
        # gdf = df.groupby(df.index.floor(time_interval))
        gdf = df.groupby(pd.Grouper(freq=time_interval))
        group = iter(gdf)
        for _ in range(len(gdf) - 1):
            interval, chunk = next(group)
            # len=0 can happen if a time gap > interval exists in data
            if len(chunk) == 0:
                continue
            yield chunk
        current_interval, next_df = next(group)
        consumed_dfs = [next_df]

    # Yield remaining chunks. Should always be at least one.
    # Can be more if the main loop is skipped (len(dataframes)=1)
    # and time_interval < time span of the single dataframe

    df = pd.concat(consumed_dfs)
    gdf = df.groupby(df.index.floor(time_interval))
    for interval, chunk in gdf:
        if len(chunk) <= 1:
            continue
        yield chunk


def multifile_read_csv(files, *args, **kwargs):
    """Buffered pd.read_csv of data split across multiple files."""
    for file_ in files:
        try:
            df = pd.read_csv(file_, *args, **kwargs)
        except Exception as e:
            mssg = "Skipping file " + str(file_) + " because " + e.args[0]
            warnings.warn(mssg, HFDataReadWarning)
            continue
        if isinstance(df, pd.io.parsers.TextFileReader):
            yield from df
        else:
            yield df


def multifile_read_ghg(files, *args, **kwargs):
    """Buffered pd.read_csv of data split across multiple files."""
    for file_ in files:
        with zipfile.ZipFile(file_) as z:
            fn = os.path.basename(file_)
            try:
                df = pd.read_csv(z.open(fn[:-3] + "data"), *args, **kwargs)
            except Exception as e:
                mssg = "Skipping file " + str(file_) + " because " + e.args[0]
                warnings.warn(mssg, HFDataReadWarning)
                continue
        if isinstance(df, pd.io.parsers.TextFileReader):
            yield from df
        else:
            yield df


def multifile_read_tob1(tobfiles, count=-1):
    """Buffered read of multiple tob1 files into dataframes."""
    for tobfile in tobfiles:
        try:
            df = dataframe_read_tob1(tobfile, count)
        except Exception as e:
            mssg = "Skipping file " + str(tobfile) + " because " + e.args[0]
            warnings.warn(mssg, HFDataReadWarning)
            continue
        yield df


def dataframe_read_tob1(tobfile, count=-1):
    return pd.DataFrame(ndarray_read_tob1(tobfile, count))


def ndarray_read_tob1(tobfile, count=-1):
    """Read TOB1 data file into structured numpy array."""
    with open(tobfile, "rb") as f:
        f.readline()
        names = f.readline().decode().strip().replace('"', "").split(",")
        f.readline()
        f.readline()
        types = f.readline().decode().strip().replace('"', "").split(",")
        dtype = np.dtype([(n, NP_TYPE[t]) for n, t in zip(names, types)])
        return np.fromfile(f, dtype=dtype, count=count)


def stats2(sarray, names=None):
    """Calculate means and (co)variances for structured array data."""

    if names is None:
        names = sarray.dtype.names
    nvar = len(names)
    data = tuple(sarray[name] for name in names)
    cov = np.cov(data)
    nondiag_cov = list(cov[i, j] for i, j in permutations(range(nvar), 2))

    names_ave = list("ave_" + name for name in names)
    names_var = list("var_" + name for name in names)
    names_cov = list(
        "cov_" + n1 + "_" + n2 for n1, n2 in permutations(names, 2)
    )

    out = dict(zip(names_ave, np.mean(data, axis=1)))
    out.update(zip(names_var, cov.diagonal()))
    out.update(zip(names_cov, nondiag_cov))

    NamedStats = namedtuple("Stats2", names_ave + names_var + names_cov)
    return NamedStats(**out)


def sat_vapor_press(t_kelvin):
    # e_sat
    tr = 1 - 373.15 / t_kelvin
    arg = 13.3185 * tr - 1.9760 * tr ** 2 - 0.6445 * tr ** 3 - 0.1299 * tr ** 4
    return 101325.0 * exp(arg)


def vapor_press_deficit(rho_vapor, t_kelvin):
    return sat_vapor_press(t_kelvin) - rho_vapor * Rgas.vapor * t_kelvin


def vapor_press_deficit_mass(rho_vapor, t_kelvin):
    return vapor_press_deficit(rho_vapor, t_kelvin) / Rgas.vapor / t_kelvin


def qflux_mass_to_heat(massflux, Tk):  # kg/m^2/s, K
    Lv = 2.5008e6 - 2366.8 * (Tk - 273.15)  # J/kg
    return massflux * Lv  # W/m^2


def cflux_mass_to_mol(massflux):  # kg/m^2/s
    return 1.0 / MW.co2 * massflux  # mol/m^2/s


def qflux_mass_to_mol(massflux):  # kg/m^2/s
    return 1.0 / MW.vapor * massflux  # mol/m^2/s


def progressive_lowcut_series(series):
    """Progressively remove low-frequency components of 1D series.

    Yields sequence in which the low-frequency (large-scale) components
    of `series` are progressively removed.  The sequence is obtained
    from reconstruction of a multilevel discrete haar wavelet
    decomposition of `series`.

    N.B.: The length of `series` is assumed to be a power of 2
    (does not check!)

    Parameters
    ----------
    series : array_like
        1D data series with a length that is a power of 2

    Yields
    -------
    lowcut_series : array
        Sequence of progressively lowcut filtered data `series`. The
        yielded series have the same length as `series`.

    Notes
    -----
    After an N level discrete wavelet decomposition, a data series S can
    be reconstructed in terms of wavelet 'approximations' (A) and
    'details' (D)::

        S = A(N) + D(N) + D(N-1) ... D(2) + D(1)
          = A(N-1) + D(N-1) + ... D(2) + D(1)     [A(N-1) = A(N) + D(N)]
            ...
          = A(1) + D(1)                           [(A(1) = A(2) + D(2)]

    where A(N) represents the 'lowest' level approximation. For
    the haar wavelet and data S having a length that is a power of 2,
    A(N) is equal to mean(S)]

    The sequence returned by this function is::

        S - A(N),
        S - A(N-1),
        S - A(N-2),
        ...,
        S - A(1)

    This sequence is computed by the equivalent::

        S - A(N),
        S - A(N) - D(N),
        S - A(N) - D(N) - D(N-1),
        ...,
        S - A(N) - D(N) - D(N-1) - ... - D(2),

    i.e. the details are removed in sequence::

        lowcut = S - A(N)
        for j = N to 2
            lowcut -= D(j)

    """

    series_data = np.asarray(series)
    wavelet = pywt.Wavelet("haar")
    nlevel = pywt.dwt_max_level(series_data.size, wavelet.dec_len)
    decomp_coef = pywt.wavedec(series_data, wavelet=wavelet, level=nlevel)
    cAn, cD = decomp_coef[0], decomp_coef[1:]
    lowcut_series = series_data - pywt.upcoef("a", cAn, wavelet, level=nlevel)
    yield lowcut_series
    for j, cDj in enumerate(cD[:-1]):
        Dj = pywt.upcoef("d", cDj, wavelet, level=nlevel - j)
        lowcut_series -= Dj
        yield lowcut_series


if __name__ == "__main__":
    pass



import os
import pickle
from types import SimpleNamespace

import numpy.testing as npt
import pandas as pd
from pandas._testing import assert_frame_equal

# from fluxpart import flux_partition
from fluxpart import fvs_partition, fpread

TESTDIR = os.path.dirname(os.path.realpath(__file__))


def test_flux_partition():
    """Integrated test of highest level flux_partition function.

    In test below, the "desired" results are not exact or even
    necessarily correct. They were obtained with an independent (matlab)
    code, and there should be reasonable agreement.
    """

    wue_data = {
        "meas_ht": 7.11,
        "canopy_ht": 4.42,
        "ppath": "C3",
        "ci_mod": "const_ppm",
        "diff_ratio": 1 / 0.7,
    }

    # soln exists for this data without any wavelet filtering
    fname = os.path.join(
        TESTDIR, "data/TOA5_6843.ts_Above_2012_06_07_1300.dat"
    )

    matlab_fluxes = SimpleNamespace(
        Fcp=-1.02807378762793,
        Fcr=0.402683101177712,
        Fqe=0.00500869036088203,
        Fqt=0.145615860044424,
        Fcp_mol=-23.3600042633021,
        Fcr_mol=9.14980916104776,
        LEe=12.1870591763138,
        LEt=354.310004313924,
    )

    fvsp = fvs_partition(fname, wue_options=wue_data, hfd_format="ec-TOA5")

    npt.assert_allclose(fvsp.df["fvsp_solution"]["var_cp"], 18.9272, atol=1)
    assert_flux_components(fvsp.df["fluxes"], matlab_fluxes)

    # test reading heights and temperature from file
    wue_data = {
        "heights": os.path.join(TESTDIR, "data/heights.csv"),
        "leaf_temper": os.path.join(TESTDIR, "data/leaf_temper.csv"),
        "ppath": "C3",
        "ci_mod": "const_ppm",
        "diff_ratio": 1 / 0.7,
    }

    fvsp = fvs_partition(fname, wue_options=wue_data, hfd_format="ec-TOA5")

    npt.assert_allclose(fvsp.df["fvsp_solution"]["var_cp"], 18.9272, atol=1)
    assert_flux_components(fvsp.df["fluxes"], matlab_fluxes)

    # soln is obtained after some wavelet filtering
    fname = os.path.join(
        TESTDIR, "data/TOA5_6843.ts_Above_2012_06_07_1245.dat"
    )

    wue_data = {
        "meas_ht": 7.11,
        "canopy_ht": 4.42,
        "ppath": "C3",
        "ci_mod": "const_ppm",
        "diff_ratio": 1 / 0.7,
    }

    matlab_fluxes = SimpleNamespace(
        Fcp=-0.866856083109642,
        Fcr=0.353428894620522,
        Fqe=0.0124697200158396,
        Fqt=0.117438136138301,
        Fcp_mol=-23.1074540435422,
        Fcr_mol=10.6590633820467,
        LEe=35.6007693518818,
        LEt=335.283229492226,
    )

    fvsp = fvs_partition(fname, wue_options=wue_data, hfd_format="ec-TOA5")

    npt.assert_allclose(
        fvsp.df.iloc[0]["fvsp_solution"]["var_cp"], 15.2944, atol=1
    )
    assert_flux_components(fvsp.df.iloc[0]["fluxes"], matlab_fluxes)


def test_fpread(tmpdir):
    """Test of writing and reading partitioning results."""

    wue_data = {
        "meas_ht": 7.11,
        "canopy_ht": 4.42,
        "ppath": "C3",
        "ci_mod": "const_ratio",
    }
    fname = os.path.join(
        TESTDIR, "data/TOA5_6843.ts_Above_2012_06_07_1300.dat"
    )
    fvsp = fvs_partition(
        fname, wue_options=wue_data, hfd_format="ec-TOA5", interval="1min"
    )
    results_file = os.path.join(tmpdir, "foo.pkl")

    fvsp.save(results_file)
    readfvsp = fpread(results_file)
    assert_frame_equal(fvsp.df, readfvsp.df)

    fvsp.save_pickle(results_file)
    readfvsp = fpread(results_file)
    assert_frame_equal(fvsp.df, readfvsp.df)

    fvsp.save_csv(results_file)
    readfvsp = fpread(results_file)
    #TODO: not testing wave_lvl, it shouldn't be a tuple
    assert_frame_equal(
        fvsp.df.drop("wave_lvl", axis=1, level=1),
        readfvsp.df.drop("wave_lvl", axis=1, level=1),
    )


def assert_flux_components(calc, desired):
    npt.assert_allclose(calc.Fcp, desired.Fcp, atol=0.5)
    npt.assert_allclose(calc.Fcr, desired.Fcr, atol=0.1)
    npt.assert_allclose(calc.Fqe, desired.Fqe, atol=0.01)
    npt.assert_allclose(calc.Fqt, desired.Fqt, atol=0.05)
    npt.assert_allclose(calc.Fcp_mol, desired.Fcp_mol, atol=10)
    npt.assert_allclose(calc.Fcr_mol, desired.Fcr_mol, atol=10)
    npt.assert_allclose(calc.LEe, desired.LEe, atol=10)
    npt.assert_allclose(calc.LEt, desired.LEt, atol=50)


if __name__ == "__main__":
    test_flux_partition()



import io
import os

import numpy as np
import numpy.testing as npt
import pandas as pd

from fluxpart.hfdata import HFData, HFDataSource
from fluxpart.fluxpart import _converter_func, _peektime

TESTDIR = os.path.dirname(os.path.realpath(__file__))
DATADIR = os.path.join(TESTDIR, "data")


def test_hfdata_reader():
    toy_data = (
        "foobar baz\n"
        "'asdf',0,2,3,4,5,6,7,9,0\n"
        "'asdf',1,2,3,4,5,6,7,9,0\n"
        "'asdf',2,2,3,4,5,6,7,9,1\n"
        "'asdf',3,2,3,4,5,6,,9,0\n"
        "'asdf',4,2,3,4,5,6,7,9,0\n"
        "# foo\n"
        "'asdf',5,2,3,4,5,6,7,9,0\n"
        "'asdf',6,2,3,4,5,6,7,xxx,0\n"
        "'asdf',7,???,3,4,5,6,7,9,0\n"
        "'asdf',8,2,3,4,5,6,7,9,0\n"
        "'asdf',9,2,3,4,5,6,7,9,0\n"
        "'asdf',10,2,3,4,5,6,7,9,0\n"
        "'asdf',11,-2,3,4,5,6,7,9,0\n"
    )

    source = HFDataSource(
        files=[io.BytesIO(toy_data.encode())],
        filetype="csv",
        cols=(1, 2, 3, 6, 7, 4, 5),
        comment="#",
        skiprows=1,
        na_values="???",
        converters={"q": _converter_func(10., 0)},
        flags=(9, 0),
        delimiter=",",
    )

    toy = HFData(next(source.reader(interval=None)))
    toy.cleanse(rd_tol=0.1, ad_tol=2, bounds={"v": (0, np.inf)})

    npt.assert_allclose(toy.dataframe["u"], [4, 5, 6])
    npt.assert_allclose(toy.dataframe["v"], 3 * [2])
    npt.assert_allclose(toy.dataframe["w"], 3 * [3])
    npt.assert_allclose(toy.dataframe["q"], 3 * [70])
    npt.assert_allclose(toy.dataframe["c"], 3 * [6])
    npt.assert_allclose(toy.dataframe["T"], 3 * [4])
    npt.assert_allclose(toy.dataframe["P"], 3 * [5])

    # missing time series data
    toy_data = (
        "foobar baz\n"
        "'2013-01-14 08:21:48'  ,0,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 08:21:48.1',1,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 08:21:48.2',2,2,3,4,5,6,7,9,1\n"
        "'2013-01-14 08:21:48.3',3,2,3,4,5,6,,9,0\n"
        "'2013-01-14 08:21:48.4',4,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10'  ,5,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.1',6,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.2',7,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.3',8,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.4',9,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.5',10,2,3,4,5,6,7,9,0\n"
        "'2013-01-14 09:20:10.6',11,-2,3,4,5,6,7,9,0\n"
    )

    source = HFDataSource(
        files=[io.BytesIO(toy_data.encode())],
        filetype="csv",
        cols=(1, 2, 3, 6, 7, 4, 5),
        time_col=0,
        skiprows=1,
        converters={"q": _converter_func(10., 0)},
        delimiter=",",
    )

    toy = HFData(next(source.reader(interval=None)))
    toy.cleanse(rd_tol=0.1, ad_tol=2)
    assert int(toy.dataframe.iloc[0]["u"]) == 6
    assert int(toy.dataframe.iloc[-1]["u"]) == 11

    # toa5
    cols = (2, 3, 4, 5, 6, 7, 8)
    fname = os.path.join(DATADIR, "TOA5_6843.ts_Above_2012_06_07_1300.dat")

    kws = dict(
        skiprows=4,
        converters={
            "T": _converter_func(1, 273.15),
            "q": _converter_func(1e-3, 0),
            "c": _converter_func(1e-6, 0),
            "P": _converter_func(1e3, 0),
        },
        flags=[(9, 0)],
    )

    source = HFDataSource(files=[fname], filetype="csv", cols=cols, **kws)
    reader = source.reader(interval=None)
    data = HFData(next(reader))
    assert_1300_read(data)

    kws = dict(
        skiprows=4,
        time_col=0,
        converters={
            "T": _converter_func(1, 273.15),
            "q": _converter_func(1e-3, 0),
            "c": _converter_func(1e-6, 0),
            "P": _converter_func(1e3, 0),
        },
        flags=(9, 0),
    )

    source = HFDataSource(files=[fname], filetype="csv", cols=cols, **kws)
    reader = source.reader(interval="15min")
    data = HFData(next(reader))
    assert data.dataframe.index[0] == pd.to_datetime("2012-06-07 13:00:00.05")
    assert data.dataframe.index[-1] == pd.to_datetime("2012-06-07 13:14:59.95")
    assert_1300_interval_read(data)

    fnames = [
        os.path.join(DATADIR, "TOA5_6843.ts_Above_2012_06_07_1245.dat"),
        os.path.join(DATADIR, "TOA5_6843.ts_Above_2012_06_07_1300.dat"),
    ]
    source = HFDataSource(fnames, filetype="csv", cols=cols, **kws)

    reader = source.reader(interval="10min")
    for cnt, df in enumerate(reader):
        assert_10min_chunk_read(cnt, df)
    reader = source.reader(interval="15min")
    for cnt, df in enumerate(reader):
        assert_15min_chunk_read(cnt, df)
    reader = source.reader(interval="20min")
    for cnt, df in enumerate(reader):
        assert_20min_chunk_read(cnt, df)

    # tob
    fname = os.path.join(DATADIR, "testing.tob")
    kws = dict(
        filetype="tob1",
        cols=(3, 4, 5, 6, 7, 8, 9),
        converters={
            "T": _converter_func(1, 273.15),
            "q": _converter_func(1e-3, 0),
            "c": _converter_func(1e-6, 0),
            "P": _converter_func(1e3, 0),
        },
    )

    source = HFDataSource(files=[fname], **kws)
    data = HFData(next(source.reader(interval="30min")))
    assert_tob_read(data)

    datetime = _peektime([fname], **kws)[0]
    assert datetime.round(freq="100ms") == pd.to_datetime(
        "2017-08-03 00:00:00.1"
    )

    reader = source.reader(interval="5s")
    for cnt, df in enumerate(reader):
        assert_5S_tobchunk_read(cnt, df)
    reader = source.reader(interval="1min")
    for cnt, df in enumerate(reader):
        assert_1min_tobchunk_read(cnt, df)

    # ghg tab delimited data file
    fname = os.path.join(DATADIR, "head_ghg.data")
    kws = {
            'filetype': 'csv',
            'sep': '\t',
            'cols': (11, 12, 13, 7, 8, 9, 10),
            'time_col': [5, 6],
            'skiprows': 8,
            'na_values': 'NAN',
            'to_datetime_kws': {"format": "%Y-%m-%d %H:%M:%S:%f"}
    }
    source = HFDataSource(files=[fname], **kws)
    data = HFData(next(source.reader(interval=None)))
    assert_ghg_read(data)


def assert_1300_read(data):
    npt.assert_allclose(data["u"].iloc[0], 0.468)
    npt.assert_allclose(data["v"].iloc[0], -0.9077501)
    npt.assert_allclose(data["w"].iloc[0], 0.1785)
    npt.assert_allclose(data["c"].iloc[0], 659.7584e-6)
    npt.assert_allclose(data["q"].iloc[0], 9.530561e-3)
    npt.assert_allclose(data["T"].iloc[0], 28.52527 + 273.15)
    npt.assert_allclose(data["P"].iloc[0], 100.1938e3)
    npt.assert_allclose(data["u"].iloc[-1], 1.3675)
    npt.assert_allclose(data["v"].iloc[-1], -0.75475)
    npt.assert_allclose(data["w"].iloc[-1], -0.1775)
    npt.assert_allclose(data["c"].iloc[-1], 658.2624e-6)
    npt.assert_allclose(data["q"].iloc[-1], 9.404386e-3)
    npt.assert_allclose(data["T"].iloc[-1], 28.35199 + 273.15)
    npt.assert_allclose(data["P"].iloc[-1], 100.1938e3)

    npt.assert_allclose(data["u"].mean(), 1.43621, atol=1e-4)
    npt.assert_allclose(data["v"].mean(), -0.634818, atol=1e-4)
    npt.assert_allclose(data["w"].mean(), 0.0619483, atol=1e-4)
    npt.assert_allclose(data["c"].mean(), 659.052e-6, atol=1e-9)
    npt.assert_allclose(data["q"].mean(), 9.56732e-3, atol=1e-7)
    npt.assert_allclose(data["T"].mean(), 28.5431 + 273.15, atol=1e-4)
    npt.assert_allclose(data["P"].mean(), 100.179e3, atol=1e0)


def assert_1300_interval_read(data):
    """last line in file is not read because belongs to next interval"""
    npt.assert_allclose(data["u"].iloc[0], 0.468)
    npt.assert_allclose(data["v"].iloc[0], -0.9077501)
    npt.assert_allclose(data["w"].iloc[0], 0.1785)
    npt.assert_allclose(data["c"].iloc[0], 659.7584e-6)
    npt.assert_allclose(data["q"].iloc[0], 9.530561e-3)
    npt.assert_allclose(data["T"].iloc[0], 28.52527 + 273.15)
    npt.assert_allclose(data["P"].iloc[0], 100.1938e3)
    npt.assert_allclose(data["u"].iloc[-1], 1.20050)
    npt.assert_allclose(data["v"].iloc[-1], -0.775)
    npt.assert_allclose(data["w"].iloc[-1], -0.1610)
    npt.assert_allclose(data["c"].iloc[-1], 658.3277e-6)
    npt.assert_allclose(data["q"].iloc[-1], 9.394239e-3)
    npt.assert_allclose(data["T"].iloc[-1], 28.35199 + 273.15)
    npt.assert_allclose(data["P"].iloc[-1], 100.1678e3)

    npt.assert_allclose(data["u"].mean(), 1.43621, atol=1e-4)
    npt.assert_allclose(data["v"].mean(), -0.634818, atol=1e-4)
    npt.assert_allclose(data["w"].mean(), 0.0619483, atol=1e-4)
    npt.assert_allclose(data["c"].mean(), 659.052e-6, atol=1e-9)
    npt.assert_allclose(data["q"].mean(), 9.56732e-3, atol=1e-7)
    npt.assert_allclose(data["T"].mean(), 28.5431 + 273.15, atol=1e-4)
    npt.assert_allclose(data["P"].mean(), 100.179e3, atol=1e0)


def assert_tob_read(data):
    npt.assert_allclose(data["u"].iloc[0], -2.57175016)
    npt.assert_allclose(data["v"].iloc[0], 1.6450001)
    npt.assert_allclose(data["w"].iloc[0], -0.12725)
    npt.assert_allclose(data["c"].iloc[0], 612.54150391e-6)
    npt.assert_allclose(data["q"].iloc[0], 13.11471748e-3)
    npt.assert_allclose(data["T"].iloc[0], 23.29580116 + 273.15)
    npt.assert_allclose(data["P"].iloc[0], 85.04070282e3)
    npt.assert_allclose(data["u"].iloc[-1], -2.4402502)
    npt.assert_allclose(data["v"].iloc[-1], 1.5402501)
    npt.assert_allclose(data["w"].iloc[-1], -0.11375)
    npt.assert_allclose(data["c"].iloc[-1], 615.627e-6)
    npt.assert_allclose(data["q"].iloc[-1], 13.200139e-3)
    npt.assert_allclose(data["T"].iloc[-1], 23.015879 + 273.15)
    npt.assert_allclose(data["P"].iloc[-1], 85.0407e3)
    assert data.dataframe.index[0].round(freq="100ms") == pd.to_datetime(
        "2017-08-03 00:00:00.1"
    )
    assert data.dataframe.index[-1].round(freq="100ms") == pd.to_datetime(
        "2017-08-03 00:00:14.4"
    )


def assert_ghg_read(data):
    npt.assert_allclose(data["u"].iloc[0], 0.752103)
    npt.assert_allclose(data["v"].iloc[0], -0.591885)
    npt.assert_allclose(data["w"].iloc[0], 0.199506)
    npt.assert_allclose(data["c"].iloc[0], 15.1805)
    npt.assert_allclose(data["q"].iloc[0], 817.852)
    npt.assert_allclose(data["T"].iloc[0], 27.6979)
    npt.assert_allclose(data["P"].iloc[0], 96.9071)
    npt.assert_allclose(data["u"].iloc[-1], 0.781697)
    npt.assert_allclose(data["v"].iloc[-1], -0.542902)
    npt.assert_allclose(data["w"].iloc[-1], 0.195424)
    npt.assert_allclose(data["c"].iloc[-1], 15.1822)
    npt.assert_allclose(data["q"].iloc[-1], 817.922)
    npt.assert_allclose(data["T"].iloc[-1], 27.6979)
    npt.assert_allclose(data["P"].iloc[-1], 96.9071)
    assert data.dataframe.index[0].round(freq="100ms") == pd.to_datetime(
        "2017-08-02 11:30:00"
    )
    assert data.dataframe.index[-1].round(freq="100ms") == pd.to_datetime(
        "2017-08-02 11:30:00.1"
    )


def assert_10min_chunk_read(ichunk, df):
    starts = [
        "2012-06-07 12:45:00.05",
        "2012-06-07 12:50:00",
        "2012-06-07 13:00:00",
        "2012-06-07 13:10:00",
    ]
    stops = [
        "2012-06-07 12:49:59.95",
        "2012-06-07 12:59:59.95",
        "2012-06-07 13:09:59.95",
        "2012-06-07 13:15:00",
    ]
    assert df.index[0] == pd.to_datetime(starts[ichunk])
    assert df.index[-1] == pd.to_datetime(stops[ichunk])


def assert_15min_chunk_read(ichunk, df):
    starts = [
        "2012-06-07 12:45:00.05",
        "2012-06-07 13:00:00",
        "2012-06-07 13:15:00",
    ]
    stops = [
        "2012-06-07 12:59:59.95",
        "2012-06-07 13:14:59.95",
        "2012-06-07 13:15:00",
    ]
    assert df.index[0] == pd.to_datetime(starts[ichunk])
    assert df.index[-1] == pd.to_datetime(stops[ichunk])


def assert_20min_chunk_read(ichunk, df):
    starts = ["2012-06-07 12:45:00.05", "2012-06-07 13:00:00"]
    stops = ["2012-06-07 12:59:59.95", "2012-06-07 13:15:00"]
    assert df.index[0] == pd.to_datetime(starts[ichunk])
    assert df.index[-1] == pd.to_datetime(stops[ichunk])


def assert_5S_tobchunk_read(cnt, df):
    starts = [
        "2017-08-03 00:00:00.1",
        "2017-08-03 00:00:05",
        "2017-08-03 00:00:10",
    ]
    stops = [
        "2017-08-03 00:00:04.9",
        "2017-08-03 00:00:09.9",
        "2017-08-03 00:00:14.4",
    ]
    assert df.index[0].round(freq="100ms") == pd.to_datetime(starts[cnt])
    assert df.index[-1].round(freq="100ms") == pd.to_datetime(stops[cnt])


def assert_1min_tobchunk_read(cnt, df):
    starts = ["2017-08-03 00:00:00.1"]
    stops = ["2017-08-03 00:00:14.4"]
    assert df.index[0].round(freq="100ms") == pd.to_datetime(starts[cnt])
    assert df.index[-1].round(freq="100ms") == pd.to_datetime(stops[cnt])


if __name__ == "__main__":
    test_hfdata_reader()



from types import SimpleNamespace
import numpy.testing as npt
from fluxpart.wue import water_use_efficiency


def test_water_use_efficiency():
    """ Unit test does not guarantee the wue function is correct, but if
    this fails, something has changed, and better understand why.
    """

    hf_stats = SimpleNamespace(
        rho_vapor=9.607e-3,
        rho_co2=658.8e-6,
        T=28.56 + 273.15,
        P=100.1e3,
        cov_w_q=0.1443e-3,
        cov_w_c=-1.059e-6,
        cov_w_T=0.1359,
        ustar=0.4179,
        rho_totair=1.150,
    )

    wue = water_use_efficiency(
        hf_stats,
        meas_ht=7.11,
        canopy_ht=4.42,
        ppath="C3",
        ci_mod="const_ppm",
        diff_ratio=(1 / 0.7),
    )

    npt.assert_allclose(wue.wue, -6.45e-3, atol=0.005e-3)
    npt.assert_allclose(wue.ambient_h2o, 12.4e-3, atol=0.05e-3)
    npt.assert_allclose(wue.ambient_co2, 638.e-6, atol=0.5e-6)
    npt.assert_allclose(wue.inter_h2o, 28.3e-3, atol=0.05e-3)
    npt.assert_allclose(wue.inter_co2, 492.e-6, atol=0.5e-6)


if __name__ == "__main__":
    test_water_use_efficiency()



import io
from unittest import TestCase

import numpy as np
import numpy.testing as npt
import pandas as pd
from pandas._testing import assert_frame_equal

from fluxpart.util import (
    stats2,
    multifile_read_csv,
    chunked_df,
    HFDataReadWarning,
)


def test_stats2():
    """Test stats2 func from fluxpart.util"""

    data = "7 8 4\n6 1 3\n10 6 6\n6 7 3\n8 2 4"
    dtype = [("v0", int), ("v1", int), ("v2", int)]
    arr = np.genfromtxt(io.BytesIO(data.encode()), dtype=dtype)
    ans = stats2(arr)

    npt.assert_allclose(ans.ave_v0, 37 / 5)
    npt.assert_allclose(ans.ave_v1, 24 / 5)
    npt.assert_allclose(ans.ave_v2, 4)
    npt.assert_allclose(ans.var_v0, 14 / 5)
    npt.assert_allclose(ans.var_v1, 97 / 10)
    npt.assert_allclose(ans.var_v2, 3 / 2)
    npt.assert_allclose(ans.cov_v0_v1, 3 / 5)
    npt.assert_allclose(ans.cov_v0_v2, 2)
    npt.assert_allclose(ans.cov_v1_v0, ans.cov_v0_v1)
    npt.assert_allclose(ans.cov_v1_v2, 1)
    npt.assert_allclose(ans.cov_v2_v0, ans.cov_v0_v2)
    npt.assert_allclose(ans.cov_v2_v1, ans.cov_v1_v2)

    data = "7 8 4\n6 1 3\n10 6 6\n6 7 3\n8 2 4"
    dtype = [("v0", int), ("v1", int), ("v2", int)]
    arr = np.genfromtxt(io.BytesIO(data.encode()), dtype=dtype)
    ans = stats2(arr, names=("v0", "v2"))

    npt.assert_allclose(ans.ave_v0, 37 / 5)
    npt.assert_allclose(ans.ave_v2, 4)
    npt.assert_allclose(ans.var_v0, 14 / 5)
    npt.assert_allclose(ans.var_v2, 3 / 2)
    npt.assert_allclose(ans.cov_v0_v2, 2)
    npt.assert_allclose(ans.cov_v2_v0, ans.cov_v0_v2)

    assert not hasattr(ans, "ave_v1")
    assert not hasattr(ans, "var_v1")
    assert not hasattr(ans, "cov_v0_v1")
    assert not hasattr(ans, "cov_v1_v0")
    assert not hasattr(ans, "cov_v1_v2")
    assert not hasattr(ans, "cov_v2_v1")


def test_mulitifile_read_csv():
    file1 = io.BytesIO("1,2,3\n4,5,6\n7,8,9\n10,11,12".encode())
    file2 = io.BytesIO("13,14,15\n16,17,18\n19,20,21\n22,23,24".encode())
    reader = multifile_read_csv([file1, file2], header=None)
    dfs = [
        pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]]),
        pd.DataFrame([[13, 14, 15], [16, 17, 18], [19, 20, 21], [22, 23, 24]]),
    ]
    for cnt, df in enumerate(reader):
        assert_frame_equal(df, dfs[cnt])

    file1 = io.BytesIO("1,2,3\n4,5,6\n7,8,9\n10,11,12".encode())
    file2 = io.BytesIO("13,14,15\n16,17,18\n19,20,21\n22,23,24".encode())
    reader = multifile_read_csv([file1, file2], header=None, chunksize=2)
    dfs = [
        pd.DataFrame([[1, 2, 3], [4, 5, 6]]),
        pd.DataFrame([[7, 8, 9], [10, 11, 12]], index=[2, 3]),
        pd.DataFrame([[13, 14, 15], [16, 17, 18]]),
        pd.DataFrame([[19, 20, 21], [22, 23, 24]], index=[2, 3]),
    ]
    for cnt, df in enumerate(reader):
        assert_frame_equal(df, dfs[cnt])

    file1 = io.BytesIO("1,2,3\n4,5,6\n7,8,9\n10,11,12".encode())
    file2 = io.BytesIO("13,14,15\n16,17,18\n19,20,21\n22,23,24".encode())
    reader = multifile_read_csv([file1, file2], header=None, chunksize=3)
    dfs = [
        pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
        pd.DataFrame([[10, 11, 12]], index=[3]),
        pd.DataFrame([[13, 14, 15], [16, 17, 18], [19, 20, 21]]),
        pd.DataFrame([[22, 23, 24]], index=[3]),
    ]
    for cnt, df in enumerate(reader):
        assert_frame_equal(df, dfs[cnt])


def test_chunked_df():
    df = pd.DataFrame(np.random.rand(50, 3))
    df = df.set_index(pd.date_range("2000-01-01", periods=50, freq="1h"))
    df1, df2 = df.iloc[:40], df.iloc[40:]
    cdf = chunked_df(iter([df1, df2]), time_interval="1D")
    df = next(cdf)
    assert df.index[0] == pd.to_datetime("2000-01-01 00:00:00")
    assert df.index[-1] == pd.to_datetime("2000-01-01 23:00:00")
    df = next(cdf)
    assert df.index[0] == pd.to_datetime("2000-01-02 00:00:00")
    assert df.index[-1] == pd.to_datetime("2000-01-02 23:00:00")
    df = next(cdf)
    assert df.index[0] == pd.to_datetime("2000-01-03 00:00:00")
    assert df.index[-1] == pd.to_datetime("2000-01-03 01:00:00")


def test_read_warning():
    # not working
    # see: https://bugs.python.org/issue29620
    tc = TestCase()
    with tc.assertWarns(HFDataReadWarning):
        hfread_error()


def hfread_error():
    file1 = io.BytesIO("1,2,3\n4,5,6\n7,8,9\n10,11,12".encode())
    file2 = io.BytesIO("13 14 15\n16,17,18\n19,20,21\n22,23,24".encode())
    reader = multifile_read_csv([file1, file2], header=None)
    for cnt, df in enumerate(reader):
        pass


if __name__ == "__main__":
    test_stats2()
    test_mulitifile_read_csv()
    test_chunked_df()
    # test_read_warning()



"""TODO:"""
import math
from types import SimpleNamespace

import numpy as np

import fluxpart.util as util
from .containers import FVSPSolution, MassFluxes, RootSoln, WQCData


class Error(Exception):
    pass


class FVSError(Error):
    def __init__(self, message):
        self.message = message


def fvspart_progressive(w, q, c, wue, adjust_fluxes=True):
    """FVS flux partitioning using high frequency eddy covariance data.

    If a valid partitioning solution is not found for the passed series
    data, low-frequency (large-scale) components are progressively
    removed from the data until either a valid solution is found or
    the series decomposition is exhausted.

    Parameters
    ----------
    w,q,c : array_like
        1D high frequency time series for vertical wind speed `w` (m/s),
        water vapor `q` (kg/m^3), and CO2 `c` (kg/m^3).
    wue : float, `wue` < 0
        leaf-level water use efficiency, kg CO2 / kg H2O.
    adjust_fluxes : bool, optional
        Indicates whether the obtained partitioned fluxes should be
        adjusted so that the totals match the original data. Default is
        True.

    Returns
    -------
    :class:`~fluxpart.containers.FVSPSolution`

    Notes
    -----
    If a valid partitioning is not found, the returned `numersoln` and
    `wqc_data` correspond to the final iteration attempted.

    """
    max_decomp_lvl = int(np.log2(w.size))
    wq_tot = np.cov((w, q))[0, 1]
    wc_tot = np.cov((w, c))[0, 1]

    # The loop progressively filters the data until a physically valid
    # partitioning is found or the loop/filter is exhausted. The first
    # iteration of progressive_lowcut removes only the mean value
    # (q'=q-<q>, etc.), so the first iteration uses the "unfiltered"
    # deviations.

    for cnt, lowcut_wqc in enumerate(_progressive_lowcut(w, q, c)):
        wave_lvl = (max_decomp_lvl - cnt, max_decomp_lvl)

        fluxes, fvsp = fvspart_series(*lowcut_wqc, wue)
        if cnt == 0:
            if fvsp.fvsp_mssg:
                mssg_for_unfiltered_data = fvsp.fvsp_mssg
            else:
                mssg_for_unfiltered_data = "NA"

        if fvsp.rootsoln.valid_root:
            if adjust_fluxes:
                fluxes = _adjust_fluxes(fluxes, wue, wq_tot, wc_tot)
                fvsp.valid_partition, fvsp.mssg = _isvalid_partition(fluxes)
            if fvsp.valid_partition:
                break

    fvsp.wave_lvl = wave_lvl
    if not fvsp.rootsoln.valid_root:
        fvsp.valid_partition = False
        fvsp.mssg = mssg_for_unfiltered_data
    if not fvsp.valid_partition:
        fluxes = MassFluxes()
        fvsp.mssg = mssg_for_unfiltered_data
    return fluxes, fvsp


def fvspart_series(w, q, c, wue, wipe_if_invalid=False):
    """FVS partition q and c fluxes using high frequency eddy cov data.

    Parameters
    ----------
    w,q,c : array_like
        1D high frequency time series for vertical wind speed `w` (m/s),
        water vapor `q` (kg/m^3), and CO2 `c` (kg/m^3).
    wue : float, `wue` < 0
        leaf-level water use efficiency, kg CO2 / kg H2O.
    wipe_if_invalid : boolean
        If True, return default (nan) values for all mass fluxes if any
        calculated fluxes violate directional requirements. Default is
        False.

    Returns
    -------
    :class:`~fluxpart.containers.FVSPSolution`,

    """
    cov = np.cov([w, q, c])
    wqc_data = WQCData(
        wq=cov[0, 1],
        wc=cov[0, 2],
        var_q=cov[1, 1],
        var_c=cov[2, 2],
        corr_qc=cov[1, 2] / math.sqrt(cov[1, 1] * cov[2, 2]),
    )
    return fvspart_interval(wqc_data, wue)


def fvspart_interval(wqc_data, wue, wipe_if_invalid=False):
    """Partition H2O and CO2 fluxes using interval averaged data values.

    Parameters
    ----------
    wqc_data : :class:`~fluxpart.containers.WQCData`
    wue : float, kg CO2 / kg H2O
        Leaf-level water use efficiency, `wue` < 0
    wipe_if_invalid : boolean
        If True, return default (nan) values for all mass fluxes if any
        calculated fluxes violate directional requirements. Default is
        False.

    Returns
    -------
    :class:`~fluxpart.containers.FVSPSolution`

    """
    try:
        rootsoln = findroot(wqc_data, wue)
    except FVSError as e:
        fvsp = FVSPSolution(
            wqc_data=wqc_data,
            valid_partition=False,
            fvsp_mssg=e.args[0],
            rootsoln=RootSoln(valid_root=False),
        )
        return MassFluxes(), fvsp

    if not rootsoln.valid_root:
        fvsp = FVSPSolution(
            wqc_data=wqc_data,
            rootsoln=rootsoln,
            valid_partition=False,
            fvsp_mssg="Invalid Root",
        )
        return MassFluxes(), fvsp

    mass_fluxes = _mass_fluxes(
        var_cp=rootsoln.var_cp,
        corr_cp_cr=rootsoln.corr_cp_cr,
        wqc_data=wqc_data,
        wue=wue,
        co2soln_id=rootsoln.co2soln_id,
    )
    isvalid, mssg = _isvalid_partition(mass_fluxes)
    if not isvalid and wipe_if_invalid:
        mass_fluxes = MassFluxes()
    fvsps = FVSPSolution(
        wqc_data=wqc_data,
        rootsoln=rootsoln,
        valid_partition=isvalid,
        fvsp_mssg=mssg,
    )
    return mass_fluxes, fvsps


def findroot(wqc_data, wue):
    """Calculate (corr_cp_cr, var_cp).

    Parameters
    ----------
    wqc_data : namedtuple or equivalent namespace
        :class:`~fluxpart.containers.WQCData`
    wue : float
        Leaf-level water use efficiency, `wue` < 0, kg CO2 / kg H2O.

    Returns
    -------
    namedtuple
        :class:`~fluxpart.containers.RootSoln`

    """
    try:
        _check_fvsp_assumptions(wqc_data, wue)
    except FVSError:
        raise

    co2soln_id = None
    sig_cr, var_cp, corr_cp_cr = np.nan, np.nan, np.nan
    valid_root = False

    var_q, var_c = wqc_data.var_q, wqc_data.var_c
    sd_q, sd_c = math.sqrt(var_q), math.sqrt(var_c)
    wq, wc = wqc_data.wq, wqc_data.wc
    corr_qc = wqc_data.corr_qc

    # wcwq = wc / wq
    # scsq = sd_c / sd_q

    # if corr_qc > 0 and ((wue < wcwq < 0) or (0 < wcwq < scsq * corr_qc)):
    #     co2soln_id = 0 # minus root
    # if corr_qc < 0:
    #     if wcwq * corr_qc < scsq < wcwq / corr_qc:
    #         co2soln_id = 1 if scsq <= corr_qc * wue else 0

    co2soln_id = 0 # minus root
    if corr_qc < 0 and sd_c / sd_q < corr_qc * wue:
        co2soln_id = 1 # plus root

    # if co2soln_id in (0, 1):
    if 1:
        valid_root = True
        mssg = ""

        numer = -2 * corr_qc * sd_c * sd_q * wq * wc
        numer += var_c * wq ** 2 + var_q * wc ** 2
        numer *= -(corr_qc ** 2 - 1) * var_c * var_q * wue ** 2
        denom = -corr_qc * sd_c * sd_q * (wc + wq * wue)
        denom += var_c * wq + var_q * wc * wue
        denom = denom ** 2

        var_cp = numer / denom

        numer = -(corr_qc ** 2 - 1) * var_c * var_q * (wc - wq * wue) ** 2
        denom = -2 * corr_qc * sd_c * sd_q * wc * wq
        denom += var_c * wq ** 2 + var_q * wc ** 2
        denom *= -2 * corr_qc * sd_c * sd_q * wue + var_c + var_q * wue ** 2

        rho_sq = numer / denom
        corr_cp_cr = -math.sqrt(rho_sq)

        wcr_ov_wcp = flux_ratio(
            var_cp, corr_cp_cr, wqc_data, "co2", co2soln_id
        )
        sig_cr = wcr_ov_wcp * math.sqrt(var_cp) / corr_cp_cr

    return RootSoln(
        corr_cp_cr=corr_cp_cr,
        var_cp=var_cp,
        sig_cr=sig_cr,
        co2soln_id=co2soln_id,
        valid_root=valid_root,
    )


def flux_ratio(var_cp, corr_cp_cr, wqc_data, ftype, farg):
    """Compute the nonstomatal:stomatal ratio of the H2O or CO2 flux.

    The ratio (either wqe/wqt or wcr/wcp) is found by solving Eq. 13
    of [SS08]_.

    Parameters
    ---------
    wqc_data : namedtuple or equivalent namespace
        :class:`~fluxpart.containers.WQCData`
    ftype : {'co2', 'h2o'}
        Specifies whether the flux is CO2 or H2O
    farg : number
        If `ftype` = 'co2', then `farg` = {0 or 1} specifies the root
        of Eq. 13b to be used to calculate the CO2 flux ratio wcr/wcp:
        `farg`=1 uses the '+' solution, `farg`=0 uses the '-' solution.
        If `ftype` = 'h2o', then `farg` is a float equal to the water
        use efficiency (wue < 0), kg/kg.

    Returns
    -------
    fratio : float or np.nan
        The requested flux ratio, wqe/wqt or wcr/wcp. Set to np.nan
        if solution is not real.

    Notes
    -----
    When solving for wqe/wqt, the '-' solution of the quadratic Eq. 13a
    is not relevant because only the '+' solution yields wqe/wqt > 0,
    which is required/assumed by the physical model in [SS08]_.

    """
    if ftype.lower() == "co2":
        sign = 1 if farg == 1 else -1
        num = wqc_data.var_c
    else:  # is "h2o"
        sign = 1
        num = farg ** 2 * wqc_data.var_q
    disc = 1 - 1 / corr_cp_cr ** 2 + num / var_cp / corr_cp_cr ** 2
    if disc < 0:
        fratio = np.nan
    else:
        fratio = corr_cp_cr ** 2 * (sign * math.sqrt(disc) - 1)
    return fratio


def _mass_fluxes(var_cp, corr_cp_cr, wqc_data, wue, co2soln_id):
    """Calculate flux components for given (var_cp, corr_cp_cr) pair."""
    wqe_ov_wqt = flux_ratio(var_cp, corr_cp_cr, wqc_data, "h2o", wue)
    wqt = wqc_data.wq / (wqe_ov_wqt + 1)
    wcp = wqt * wue
    wcr = wqc_data.wc - wcp
    wqe = wqc_data.wq - wqt
    return MassFluxes(
        Fq=wqc_data.wq, Fqt=wqt, Fqe=wqe, Fc=wqc_data.wc, Fcp=wcp, Fcr=wcr
    )


def _check_fvsp_assumptions(qcdat, wue):
    pqc = qcdat.corr_qc
    wcwq = qcdat.wc / qcdat.wq
    scsq = np.sqrt(qcdat.var_c / qcdat.var_q)
    mssg = ""
    if wue > wcwq:
        mssg += "wue>Fc/Fq; ".format(wue, wcwq)
    if wcwq >= pqc * scsq:
        m = "Fc/Fq>=pqc*sigc/sigq; "
        mssg += m.format(wcwq, scsq * pqc)
    if pqc < 0 and wcwq < scsq / pqc:
        m = "Fc/Fq<sigc/sigq/pqc; "
        mssg += m.format(wcwq, scsq / pqc)
    if math.isclose(pqc, 0, abs_tol=1e-15):
        mssg += "pqc=0;".format(pqc)
    if mssg:
        # vals = "(Fc/Fq={:.4}; sigc/sigq={:.4}; W={:.4}; pqc={:.4})"
        # vals = vals.format(wcwq, scsq, wue, pqc)
        raise FVSError(mssg) # + vals)


def _isvalid_partition(flux_components):
    """Test if partitioned flux directions (signs) are valid."""
    fc = flux_components
    isvalid = True
    mssg = ""
    if fc.Fqt <= 0:
        isvalid = False
        mssg += "Fqt <= 0; "
    if fc.Fqe <= 0:
        isvalid = False
        mssg += "Fqe <= 0; "
    if fc.Fcp >= 0:
        isvalid = False
        mssg += "Fcp >= 0; "
    if fc.Fcr <= 0:
        isvalid = False
        mssg += "Fcr <= 0; "
    return isvalid, mssg or "OK"


def _adjust_fluxes(flux_components, wue, Fq_tot, Fc_tot):
    """Adjust partitioned fluxes so they match measured totals.

    If filtering has been applied to the series data, covariances in
    the filtered data may differ from those in the original data.
    Consequently, partitioned flux totals may not match exactly the
    total fluxes indicated by the original data. Here, partitioned
    fluxes are adjusted proportionally so that they match the totals in
    the original data.

    Parameters
    ----------
    flux_components : namedtuple or equivalent
        Attributes (floats) specify the mass flux components (Fq, Fqe,
        Fq, Fc, Fcr, Fcp), kg/m^2/s.
    wue : float
        Leaf-level water use efficiency (`wue` < 0), kg CO2 / kg H2O
    Fq_tot, Fc_tot : float
        Desired net total H2O (`Fq_tot`) and CO2 (`Fc_tot`) fluxes,
        kg/m^2/s.

    Returns
    -------
    namedtuple
        :class:`~fluxpart.containers.MassFluxes`

    """
    fc = flux_components
    Fq_diff = Fq_tot - (fc.Fqe + fc.Fqt)
    Fqe = fc.Fqe + Fq_diff * (fc.Fqe / (fc.Fqt + fc.Fqe))
    Fqt = Fq_tot - Fqe
    Fcp = wue * Fqt
    Fcr = Fc_tot - Fcp
    return MassFluxes(Fq=Fq_tot, Fqt=Fqt, Fqe=Fqe, Fc=Fc_tot, Fcp=Fcp, Fcr=Fcr)


def _progressive_lowcut(wind, vapor, co2):
    """Apply progressive lowcut filter to wind, vapor, and CO2 series.

    Use wavelet decomposition to yield a sequence of (w, q, c) series
    in which low frequency (large scale) components are progressively
    removed from w, q, c.

    Parameters
    ----------
    wind,vapor,co2 : array-like
        1D time series of vertical wind velocity (w), water vapor
        concentration (q), and carbon dioxide concentration (c).

    Yields
    ------
    (lc_w, lc_q, lc_c) : tuple of arrays
        Arrays lc_w, lc_q, and lc_c are low cut (high pass) filtered
        versions of the passed w,q,c series.

    Notes
    -----
    Before the filter is applied, the data are truncated so that the
    length is a power of 2.

    """
    max_pow2_len = 2 ** int(np.log2(np.asarray(co2).shape[0]))
    trunc_w = np.asarray(wind)[:max_pow2_len]
    trunc_q = np.asarray(vapor)[:max_pow2_len]
    trunc_c = np.asarray(co2)[:max_pow2_len]
    lowcut_w = util.progressive_lowcut_series(trunc_w)
    lowcut_q = util.progressive_lowcut_series(trunc_q)
    lowcut_c = util.progressive_lowcut_series(trunc_c)
    for lowcut_series in zip(lowcut_w, lowcut_q, lowcut_c):
        yield lowcut_series



import os
from setuptools import setup

HERE = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(HERE, "fluxpart", "__version__.py")) as f:
    exec(f.read(), about)

LONG = (
    "Python 3 module implementing the Scanlon and Sahu (2008) procedure "
    "for partitioning eddy covariance measurements of water vapor and "
    "carbon dioxide fluxes into stomatal (transpiration, photosynthesis) "
    "and nonstomatal (evaporation, respiration) components."
)

SHORT = "Module for partitioning eddy covariance flux measurements."

setup(
    name="fluxpart",
    version=about["__version__"],
    description=SHORT,
    long_description=LONG,
    url="https://github.com/usda-ars-ussl/fluxpart",
    author="Todd Skaggs",
    author_email="todd.skaggs@ars.usda.gov",
    license="Public Domain CC0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["fluxpart"],
    install_requires=["numpy", "matplotlib", "pandas", "pywavelets", "attrs"],
    zip_safe=False,
    tests_require=["numpy", "pandas"],
    test_suite="tests",
    extras_require={"tests": ["pytest"], "docs": ["sphinx >= 1.4"]},
)



from math import log, sqrt

from .constants import GRAVITY, VON_KARMAN
from .constants import MOLECULAR_WEIGHT as MW
from .constants import SPECIFIC_GAS_CONSTANT as Rgas
from .containers import WUE
from .util import (
    sat_vapor_press,
    vapor_press_deficit,
    vapor_press_deficit_mass,
)


_C3_DEFAULTS = dict(
    const_ppm=280.0,  # ppm
    const_ratio=0.7,
    linear=(1, 1.6e-4),
    sqrt=22e-9,  # kg-co2 / m^3 / Pa
)

_C4_DEFAULTS = dict(const_ppm=130.0, const_ratio=0.44, linear=(1, 2.7e-4))

CI_DEFAULT_PARAMS = dict(C3=_C3_DEFAULTS, C4=_C4_DEFAULTS)


class Error(Exception):
    pass


class WUEError(Error):
    def __init__(self, message):
        self.message = message


def water_use_efficiency(
    hfs,
    ci_mod,
    ci_mod_param=None,
    leaf_temper=None,
    leaf_temper_corr=0,
    meas_ht=None,
    canopy_ht=None,
    ppath=None,
    diff_ratio=1.6,
):

    """Estimate leaf-level water use efficiency.

    Parameters
    ----------
    hfs : :class:`~fluxpart.hfdata.HFSummary` namedtuple
        Summary statistics for high frequency eddy covariance data interval.
        Possesses the following attributes (all floats):
        `rho_vapor`, mean vapor density (kg/m^3);
        `rho_co2`, mean carbon dioxide concentration (kg/m^3);
        `T`, mean air temperature (K);
        `P`, mean atmospheric pressure (Pa);
        `cov_w_q`, covariance of vert wind velocity and H2O density (kg/m^2/s);
        `cov_w_q`, covariance of vert wind velocity and CO2 density (kg/m^2/s);
        `cov_w_T`, covariance of wind and temperature (K m/s);
        `ustar`, friction velocity (m/s);
        `rho_totair`, moist air density (kg/m^3).
    meas_ht : float or callable, optional
        Eddy covariance measurement height (m).
    canopy_ht : float or callable, optional
        Vegetation canopy height (m).
    ppath : {'C3', 'C4'}
        photosynthetic pathway
    ci_mod : {'const_ratio', 'const_ppm', 'linear', 'sqrt', 'opt'}
        Specifies the model to be used to determine the leaf
        intercellular CO2 concentration. See Notes below for model
        descriptions.
    ci_mod_param : float or 2-tuple of floats, optional
        Paramter values to be used with `ci_mod`. The number of
        parameters required depends on the model (see Notes). The
        default is ci_mod_param=None, in which case default values are
        used.
    leaf_temper : float, optional
        Measured canopy tempeature in degrees K. If None (defalut), the
        leaf temperature is taken to be equal to the air temperature in
        `hfs`.
    leaf_temper_corr : float, optional
        Optional adjustment to leaf temperature. The temperature used to
        calculate intercelluar vapor and CO2 concentrations is
        leaf_T + `leaf_temp_corr`, where leaf_T is `leaf_temper` if
        provided, and the air temperature in `hfs` otherwise. Default is
        `leaf_temper_corr` = 0.
    diff_ratio: float, optional
        Ratio of molecular diffusivities for water vapor and CO2.
        Default is `diff_ratio` = 1.6.

    Returns
    -------
    namedtuple
        :class:`~fluxpart.containers.WUE`

    Notes
    -----
    Leaf-level water use efficiency is estimated as [CN98]_ (pg. 238)::

        wue = (1 / DR) * (ca - ci) / (qa - qi)

    where::

        DR = `diff_ratio`
        ca, ci = ambient and intercellular CO2 concentration, resp.
        qa, qi = ambient and intercellular H2O concentration, resp.

    ca and qa are estimated from above-canopy tower measurements by
    extrapolating a logarithmic mean profile with stability corrections
    to the zero-plane displacement height [SS08]_.

    qi corresponds to 100 percent relative humidity at `leaf_temper`.

    To estimate ci, the following models are available:

    'const_ppm'
        ci (kg/m^3) is determined from a specified constant ppm value,
        ci_ppm::

            ci = f(ci_ppm; temperature, pressure)

        Default parameter values for ci_ppm are 280 ppm when ppath='C3',
        and 130 ppm when ppath='C4'. [CN98]_ (pg. 237).

    'const_ratio'
        The ratio ci/ca is constant::

            ci/ca = K.

        Default parameter values are K=0.7 for C3 plants and K=0.44 for
        C4.

    'linear'
        ci/ca is a linear function of vpd (the atmospheric vapor
        pressure deficit, which is calculated internally)::

            ci/ca = b - m * vpd

        b is dimensionless with a value of ~1 while m has units of 1/Pa.
        The parameter pair (b, m) defaluts to (1, 1.6e-4) for C3 plants
        and (1, 2.7e-4) for C4. See [MG83]_.

    'sqrt'
        ci/ca is a function of sqrt(vpd/ca) [KPO09]_::

            ci/ca = 1 - sqrt(1.6 * lambd *  vpd / ca)

        The paramater lambd has units of kg-CO2 / m^3 / Pa, and defaults
        to 22e-9 for C3 plants. The sqrt model is not enabled for C4
        plants.

    'opt'
        ci/ca is determined based on an optimization model. The
        estimated value is given by Eq. 19 of [SSS19]_. With this
        approach, no paramters (`ci_mod_param`) are specified. The model
        may not be suitable for C4 vegetation.   

    """
    if canopy_ht > meas_ht:
        raise WUEError("canopy_ht is less than meas_ht")

    # Assume zero-plane and roughness params for vapor and CO2 are the same.
    # d0 = Zero-plane displacement height (L), Eq. 5.2 of [CN98]
    d0 = 0.65 * canopy_ht
    # zv = Roughness parameter (L), Eqs. 5.3 and 7.19 of [CN98]
    zv = 0.2 * (0.1 * canopy_ht)

    # Virtual temperature flux
    Qflux = hfs.cov_w_T + 0.61 * hfs.T * hfs.cov_w_q / hfs.rho_totair  # K m/s
    # Obukhov length
    obukhov_len = -hfs.T * hfs.ustar ** 3 / (VON_KARMAN * GRAVITY * Qflux)  # m

    # Stability correction
    zeta = (meas_ht - d0) / obukhov_len
    # Unstable
    if zeta < -0.04:
        psi_v = 2.0 * log((1 + (1 - 16.0 * zeta) ** 0.5) / 2)
    # Neutral
    elif zeta <= 0.04:
        psi_v = 0.0
    # Stable
    else:
        psi_v = -5.0 * zeta

    # Ambient concentrations (kg/m^3)
    arg = (log((meas_ht - d0) / zv) - psi_v) / VON_KARMAN / hfs.ustar
    ambient_h2o = hfs.rho_vapor + hfs.cov_w_q * arg
    ambient_co2 = hfs.rho_co2 + hfs.cov_w_c * arg

    # Intercellular saturation vapor pressure `esat`
    leaf_T = (leaf_temper or hfs.T) + leaf_temper_corr
    esat = sat_vapor_press(leaf_T)

    # Intercellular vapor density.
    eps = MW.vapor / MW.dryair
    inter_h2o = hfs.rho_totair * eps * esat / (hfs.P - (1 - eps) * esat)

    if ci_mod == "opt":
        varq = hfs.var_vapor
        varc = hfs.var_co2
        wq = hfs.cov_w_q
        wc = hfs.cov_w_c
        pqc = hfs.corr_q_c
        dr = diff_ratio
        vpdm = vapor_press_deficit_mass(ambient_h2o, leaf_T)
        vpd = vapor_press_deficit(ambient_h2o, leaf_T)
        if vpdm < 0:
            _bad_vpdm = "vpd = {:.4} kg/m^3 <= 0"
            raise WUEError(_bad_vpdm.format(vpdm))
        m = -(varc * wq - pqc * sqrt(varq * varc) * wc)
        m /= varq * wc - pqc * sqrt(varq * varc) * wq
        if m < 0:
            raise WUEError("opt wue m arg = {:.4} < 0".format(m))
        arg = dr * vpdm * m
        rootarg = arg * (ambient_co2 + arg)
        if rootarg < 0:
            raise WUEError("opt wue root arg = {:.4} < 0".format(rootarg))
        wue = (arg - sqrt(rootarg)) / (dr * vpdm)
        return WUE(
            wue=wue,
            inter_h2o=inter_h2o,
            ambient_h2o=ambient_h2o,
            ambient_co2=ambient_co2,
            vpd=vpd,
            meas_ht=meas_ht,
            leaf_temper=leaf_T,
            ci_mod=ci_mod,
            diff_ratio=diff_ratio,
        )

    if meas_ht is None or canopy_ht is None or ppath is None:
        mssg = "meas_ht, canopy_ht, and ppath are required args for ci_mod= {}"
        raise WUEError(mssg.format(ci_mod))

    # Intercellular CO2 concentration, aka ci (kg/m^3)
    if isinstance(ci_mod, str):
        ci_mod = (ci_mod, None)

    # Ambient vapor pressure deficit
    # TODO should this use the extrapolated ambient h2o?
    vpd = vapor_press_deficit(ambient_h2o, leaf_T)
    if vpd < 0:
        raise WUEError("Negative vapor pressure deficit {:.4} Pa".format(vpd))

    ci_mod_name = ci_mod[0]
    if ci_mod_name == "sqrt" and ppath == "C4":
        err = "Combination of 'sqrt' ci model and 'C4' ppath not enabled"
        raise WUEError(err)
    ci_mod_params = ci_mod[1] or CI_DEFAULT_PARAMS[ppath][ci_mod_name]

    ci_dispatch = {
        "const_ppm": _ci_const_ppm(hfs.P, leaf_T, Rgas.co2),
        "const_ratio": _cica_const_ratio(ambient_co2),
        "linear": _cica_linear(ambient_co2, vpd),
        "sqrt": _cica_sqrt(ambient_co2, vpd),
    }
    inter_co2 = ci_dispatch[ci_mod_name](ci_mod_params)

    coef = 1.0 / diff_ratio
    wue = coef * (ambient_co2 - inter_co2) / (ambient_h2o - inter_h2o)

    if ambient_co2 <= inter_co2:
        mssg = "WUE estimate, ci={}, ca={}".format(ambient_co2, inter_co2)
        raise WUEError(mssg)
    if ambient_h2o >= inter_h2o:
        mssg = "WUE estimate, qi={}, qa={}".format(ambient_h2o, inter_h2o)
        raise WUEError(mssg)

    return WUE(
        wue=wue,
        inter_h2o=inter_h2o,
        inter_co2=inter_co2,
        ambient_h2o=ambient_h2o,
        ambient_co2=ambient_co2,
        vpd=vpd,
        ppath=ppath,
        meas_ht=meas_ht,
        leaf_temper=leaf_T,
        canopy_ht=canopy_ht,
        ci_mod=ci_mod_name,
        ci_mod_param=ci_mod_params,
        diff_ratio=diff_ratio,
    )


def _ci_const_ppm(pressure, temperature, Rco2):
    """ci is a fixed ppm value."""

    def ci_func(ci_ppm):
        """Return ci = intercellular CO2 concentration, kg/m^3."""
        return ci_ppm * 1e-6 * pressure / Rco2 / temperature

    return ci_func


def _cica_const_ratio(ambient_co2):
    """ci/ca is constant."""

    def ci_func(const):
        """Return ci = intercellular CO2 concentration, kg/m^3."""
        return const * ambient_co2

    return ci_func


def _cica_linear(ambient_co2, vpd):
    """ci/ca is a decreasing linear function of vapor pressure deficit."""

    def ci_func(linear_params):
        """Return ci = intercellular CO2 concentration, kg/m^3."""
        # b is unitless with a value of ~1, and m (> 0) has units of Pa^-1
        b, m = linear_params
        return ambient_co2 * (b - m * vpd)

    return ci_func


def _cica_sqrt(ambient_co2, vpd):
    """ci/ca is a function of sqrt(`vpd`/ca)."""

    def ci_func(lambd):
        """Return ci = intercellular CO2 concentration, kg/m^3."""
        # lambd has units of kg-co2 / m^3 / Pa
        return ambient_co2 * (1 - sqrt(1.6 * lambd * vpd / ambient_co2))

    return ci_func



import pickle
import pandas as pd
from .fluxpart import fvspart, FluxpartResult


def fpread(saved_results):
    try:
        return FluxpartResult(pd.read_pickle(saved_results))
    except pickle.UnpicklingError:
        df = pd.read_csv(
            saved_results, index_col=[0], header=[0, 1], parse_dates=True
        )
        return FluxpartResult(df)


def fvs_partition(
    file_or_dir,
    time_sorted=False,
    interval=None,
    hfd_format=None,
    hfd_options=None,
    meas_wue=None,
    wue_options=None,
    part_options=None,
    label=None,
    stdout=False,
    verbose=True,
):
    """Partition CO2 & H2O fluxes into stomatal & nonstomatal components.

    Provides a full implementation of the flux variance similarity
    partitioning algorithm; reads high frequency eddy covariance data;
    performs data transformations and data QA/QC; analyzes water vapor
    and carbon dioxide fluxes; and partitions the fluxes into stomatal
    (transpiration, photosynthesis) and nonstomatal (evaporation,
    respiration) components.

    The following notation is used in variable naming and documentation
    to represent meteorological quantities::

        u, v, w = wind velocities
        q = water vapor mass concentration
        c = carbon dioxide mass concentration
        T = air temperature
        P = total air pressure

    Parameters
    ----------
    file_or_dir : str or list of str
        Pathname for high frequency eddy covariance data file, data
        directory, or list of either. Paths can be relative or absolute,
        and can include unix-style wildcards and brace expansion.  If
        pathname is a directory without any file pattern, then all files
        in the directory are read.
    time_sorted : bool, optional
        Indicates whether the list of data files given by `file_or_dir`
        is sorted by time. Default is False.
    interval : str, optional
        Time interval used to aggregate data and partition fluxes.
        Specified using the pandas string alias_ format (e.g.,
        ``interval="30min"``). Set to None to treat whole, individual
        data files as the analysis interval (e.g., if data files each
        hold 30 min of data). Default is None.
    hfd_format : dict or {"ec-TOB1", "ec-TOA5"}, optional
        Dictionary of parameters specifying the high frequency data file
        format. See ``Other parameters`` below for an explanation of
        required and optional formatting parameters. Some pre-defined
        file formats can be specified by passing a string instead of a
        dictionary.  Currently two formats are defined: "ec-TOB1" and
        "ec-TOA5".  These formats correspond to files commonly created
        when using Campbell Scientific eddy covariance data logger
        software. See ``Notes`` below for an explanation of these
        pre-defined formats. Default is "ec-TOB1".
    hfd_options : dict, optional
        Dictionary of parameters specifying options for correcting
        high-frequency eddy covariance data and applying quality control
        measures. See ``Other parameters`` for a listing of options.
    meas_wue : float or callable, optional
        Measured (or otherwise prescribed) leaf-level water use
        efficiency (kg CO2 / kg H2O). Note that by definition,
        `meas_wue` must be a negative value (< 0). If callable, it
        should take a datetime object as its sole argument and return a
        float value.
    wue_options : dict, required if `meas_wue` is not provided
        Dictionary of parameters and options used to estimate water use
        efficiency if `meas_wue` is not provided. See ``Other
        parameters`` section for a description of valid fields.  When
        passing `wue_options`, it is always required to provide entries
        "canopy_ht", "meas_ht", and "ppath". Other entries are optional.
        The values for "canopy_ht" and "meas_ht" can be either a float
        or a callable. A callable must accept a datetime object as its
        sole argument and return a float value.
    part_options : dict, optional
        Dictionary of options for the fvs partitioning procedure. See
        ``Other parameters`` section for a listing of valid options.
    label : optional
        Optional identifier to be appended to the results.
    stdout : bool, optional
        If True (default), print to stdout information about the
        progress of the analysis.
    verbose : bool, optional
        If True (default), print extra progress information to stdout.


    Returns
    -------
    :class:`~fluxpart.fluxpart.FluxpartResult`


    Other Parameters
    ----------------
    hfd_format : dict
        High frequency data file format options. NOTE: When passing
        hfd_format, it is required at a minimum to provide values for
        "filetype" "cols", and "time_col" (detailed below).
        "unit_convert" and "temper_unit" are also required if data are
        not in SI units.
    hfd_format["filetype"] : {"csv", "tob1", "ghg", "pd.df"}
        "csv" = delimited text file; "tob1" = Campbell Scientific binary
        data table format; "ghg" = LI-COR raw data format; "pd.df" =
        pandas dataframe.
    hfd_format["cols"] : 7*(int,)
        7-tuple of integers indicating the data column numbers that
        contain series data for (u, v, w, c, q, T, P), in that order.
        Uses 0-based indexing.
    hfd_format["time_col"] : int or [int, int]
        The column index for the datetime column, or if a list, the
        indices for the date and time columns.

    hfd_format["unit_convert"] : dict
        Dictionary of multiplication factors required to convert any u,
        v, w, c, q, or P data not in SI units to SI units (m/s, kg/m^3,
        Pa). (Note T is not in that list). The dictionary keys are the
        variable names. For example, if all data are in SI units except
        P and c, which are in kPa and mg/m^3, respectively, then set:
        ``hfd_options["unit_convert"] = {"P": 1e3, "c": 1e-6}``,
        since it is necessary to multiply the kPa pressure data by 1e3
        to obtain the SI pressure unit (Pa), and the mg/m^3 CO2 data by
        1e-6 to obtain the SI concentration unit (kg/m^3). c and q data
        can be converted this way if they have units of mass or moles
        per volume. If instead the c and q data are given as mole
        fractions (e.g., mmol of q per mol dry air), then set the
        c and q converters to one of the following: "ppm_dry", "ppt_dry",
        "ppm_wet", or "ppt_wet". Use the "_dry" versions if the mole
        fraction data are w.r.t. to dry air (no water vapor) and "_wet"
        if they are w.r.t. moist air. For example, if q is in mmol q per
        mol dry air and c is umol c per mol dry air, then:
        ``{"P": 1e3, "q": "ppt_dry", "c": "ppm_dry"}``.
    hfd_format["temper_unit"] : {"K", "C"}
        Temperature data units. Default is "K" (Kelvin).
    hfd_format["flags"] : 2-tuple or list of 2-tuples
        Specifies that one or more data columns are used to flag bad
        data records. Each tuple is of the form (col, goodval), where
        col is an int specifying the column number containing the flag
        (0-based indexing), and goodval is the value of the flag that
        indicates a good data record.
    hfd_format["to_datetime_kws"] : dict
        Dict of keyword arguments passed to pandas.to_datetime_ to
        read datafile dates and time. Generally needed only if a
        nonstandard format is used in the datafile.
    hfd_format[ other keys ]
        when `hfd_format["filetype"]` is "csv" or "ghg", all other
        key:value pairs in `hfd_format` are passed as keyword arguments
        to pandas.read_csv_. Those keywords may be required to specify
        the details of the file formatting. Among the most commonly
        required are: "sep", the str that is used to separate values or
        define column widths (default is sep=","); and "skiprows", which
        will be needed if the file contains header rows. See
        pandas.read_csv_ for a full description of available format
        options.

    hfd_options: dict
        Options for pre-processing high frequency data.
    hfd_options["correct_external"] : bool, optional
        If True (default), the water vapor and carbon dioxide series
        data are corrected for external fluctuations associated with air
        temperature and vapor density according to [WPL80]_ and [DK07]_.
    hfd_options["bounds"] : dict, optional
        Dictionary indicating any lower and upper bounds for valid data.
        Dictionary entries have the form ``varname: (float, float)``,
        where varname is one of "u", "v", "w", "q", "c", "T", or "P",
        and the 2-tuple holds values for the lower and upper bounds:
        ``(lower, upper)``.  Data records are rejected if a variable in
        the record is outside the prescribed bounds. Default is ``bounds
        = {"c": (0, np.inf), "q": (0, np.inf)}`` such that data records
        are rejected if c or q data are not positive values.
    hfd_options["rd_tol"] : float, optional
        Relative tolerance for rejecting the datafile. Default is
        `hfd_options["rd_tol"]` = 0.4. See hfd_options["ad_tol"] for
        further explanation.
    hfd_options["ad_tol"] : int, optional
        Absolute tolerance for rejecting the datafile. Default is
        `ad_tol` = 1024. If the datafile contains bad records (not
        readable, out-of-bounds, or flagged data), the partitioning
        analysis is performed using the longest stretch of consecutive
        good data records found, unless that stretch is too short, in
        which case the analysis is aborted. The criteria for judging
        "too short" can be specified in both relative and absolute
        terms: the datafile is rejected if the good stretch is a
        fraction of the total data that is less than `rd_tol`, and/or is
        less than `ad_tol` records long.
    hfd_options["ustar_tol"] : float
        If the friction velocity (m/s) determined from the high
        frequency data is less than `ustar_tol`, the
        partitioning analysis is aborted due to insufficient turbulence.
        Defalult is `hfd_options["ustar_tol"]` = 0.1 (m/s).

    wue_options: dict
        Parameters for estimating water use efficiency. "canopy_ht",
        "meas_ht", and "ppath" are required keys, unless "heights" is
        given (in which case "canopy_ht" and "meas_ht" are ignored).
        See: :func:`~fluxpart.wue.water_use_efficiency` for more about
        water use efficiency estimation.
    wue_options["canopy_ht"] : float or callable
        Vegetation canopy height (m). If callable must accept a date
        object as its sole argument and return a float value.
    wue_options["meas_ht"] : float or callable
        Eddy covariance measurement height (m). If callable must accept
        a date object as its sole argument and return a float value.
    wue_options["heights"] : str or callable, optional
        Alternative way to specify (canopy_ht, meas_ht). If str, is the
        name of a csv file with date strings in the first column,
        canopy heights in the second, and measurement heights in the
        third. If callable, accepts date as its sole argument and
        returns the tuple of heights.
    wue_options["ppath"] : {"C3", "C4"}
        Photosynthetic pathway.
    wue_options["ci_mod"] : str
        Valid values: "const_ratio", "const_ppm", "linear", "sqrt",
        "opt".  See: :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options["ci_mod_param"] : float or (float, float)
        Paramter values to be used with `ci_mod`.
        See: :func:`~fluxpart.wue.water_use_efficiency`.
    wue_options["leaf_temper"] : float, str, or callable
        Canopy leaf temperature. The units (K or C) are whatever is
        indicated in hfd_format["temper_unit"]. If not specified, the
        temperature is set to the air temperature. If a str, is the name
        of a csv file with datetime stamps in the first column and
        temperatures in the second. If callable, must accept a datetime
        as its sole argument and return a float value.
    wue_options["leaf_temper_corr"] : float
        Offset adjustment applied to canopy temperature (C). Default is
        zero.
    wue_options["diff_ratio"]: float, optional
        Ratio of molecular diffusivities for water vapor and CO2.
        Default is `diff_ratio` = 1.6.

    part_options : dict
        Options for the fvs partitioning algorithm
    part_options["adjust_fluxes"] : bool
        If True (default), the final partitioned fluxes are adjusted
        proportionally such that sum of the partitioned fluxes match
        exactly the total fluxes indicated in the original data.
    part_options["daytime"] : 2-tuple, str, or callable, optional
        A 2-tuple of python datetime.time objects or timestamps for
        (sunrise, sunset). If specified, all fluxes for time intervals
        ending before sunrise or starting after sunset will be assumed
        non-stomatal. If callable, function should take a datetime.date
        object as its sole argument and return a 2-tuple of time objects
        or stamps. If str, is the name of a csv file with date stamp in
        the first column, sunrise time in the second, and sunset time in
        the third.

    NOTES
    -----
    Three pre-defined hfd_formats are available.

    "EC-TOA5"::

        hfd_format = {
            "filetype": "csv",
            "skiprows": 4,
            "cols": (2, 3, 4, 5, 6, 7, 8),
            "time_col": 0,
            "temper_unit": "C",
            "unit_convert": {"q": 1e-3, "c": 1e-6, "P": 1e3},
            "na_values": "NAN",
        }

    "EC-TOB1"::

        hfd_format = {
            "filetype": "tob1",
            "cols": (3, 4, 5, 6, 7, 8, 9),
            "temper_unit": "C",
            "unit_convert": {"q": 1e-3, "c": 1e-6, "P": 1e3},
        }

    "EC-GHG1"::

        hfd_format = {
            "filetype": "ghg",
            "sep": "\t",
            "cols": (11, 12, 13, 7, 8, 9, 10),
            "time_col": [5, 6],
            "unit_convert": dict(q=1e-3 * MW.vapor, c=1e-3 * MW.co2, P=1e3),
            "temper_unit": "C",
            "skiprows": 8,
            "na_values": "NAN",
            "to_datetime_kws": {"format": "%Y-%m-%d %H:%M:%S:%f"},
        }


    .. _pandas.read_csv:
        https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html

    .. _alias:
        http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

    .. _pandas.to_datetime:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.to_datetime.html

    """
    return fvspart(
        file_or_dir,
        time_sorted,
        interval,
        hfd_format,
        hfd_options,
        meas_wue,
        wue_options,
        part_options,
        label,
    )
